import numpy as np
import copy
import pybullet_utils.mpi_util as MPIUtil
from pybullet_utils.logger import Logger


class Normalizer(object):
  CHECK_SYNC_COUNT = 50000  # check synchronization after a certain number of entries

  # these group IDs must be the same as those in CharController.h
  NORM_GROUP_SINGLE = 0
  NORM_GROUP_NONE = -1

  class Group(object):

    def __init__(self, id, indices):
      self.id = id
      self.indices = indices
      return

  def __init__(self, size, groups_ids=None, eps=0.02, clip=np.inf):
    self.eps = eps
    self.clip = clip
    self.mean = np.zeros(size)
    self.mean_sq = np.zeros(size)
    self.std = np.ones(size)
    self.count = 0
    self.groups = self._build_groups(groups_ids)

    self.new_count = 0
    self.new_sum = np.zeros_like(self.mean)
    self.new_sum_sq = np.zeros_like(self.mean_sq)
    return

  def record(self, x):
    size = self.get_size()
    is_array = isinstance(x, np.ndarray)
    if not is_array:
      assert (size == 1)
      x = np.array([[x]])

    assert x.shape[-1] == size, \
        Logger.print2('Normalizer shape mismatch, expecting size {:d}, but got {:d}'.format(size, x.shape[-1]))
    x = np.reshape(x, [-1, size])

    self.new_count += x.shape[0]
    self.new_sum += np.sum(x, axis=0)
    self.new_sum_sq += np.sum(np.square(x), axis=0)
    return

  def update(self):
    new_count = MPIUtil.reduce_sum(self.new_count)
    new_sum = MPIUtil.reduce_sum(self.new_sum)
    new_sum_sq = MPIUtil.reduce_sum(self.new_sum_sq)

    new_total = self.count + new_count
    if (self.count // self.CHECK_SYNC_COUNT != new_total // self.CHECK_SYNC_COUNT):
      assert self.check_synced(), Logger.print2('Normalizer parameters desynchronized')

    if new_count > 0:
      new_mean = self._process_group_data(new_sum / new_count, self.mean)
      new_mean_sq = self._process_group_data(new_sum_sq / new_count, self.mean_sq)
      w_old = float(self.count) / new_total
      w_new = float(new_count) / new_total

      self.mean = w_old * self.mean + w_new * new_mean
      self.mean_sq = w_old * self.mean_sq + w_new * new_mean_sq
      self.count = new_total
      self.std = self.calc_std(self.mean, self.mean_sq)

      self.new_count = 0
      self.new_sum.fill(0)
      self.new_sum_sq.fill(0)

    return

  def get_size(self):
    return self.mean.size

  def set_mean_std(self, mean, std):
    size = self.get_size()
    is_array = isinstance(mean, np.ndarray) and isinstance(std, np.ndarray)

    if not is_array:
      assert (size == 1)
      mean = np.array([mean])
      std = np.array([std])

    assert len(mean) == size and len(std) == size, \
        Logger.print2('Normalizer shape mismatch, expecting size {:d}, but got {:d} and {:d}'.format(size, len(mean), len(std)))

    self.mean = mean
    self.std = std
    self.mean_sq = self.calc_mean_sq(self.mean, self.std)
    return

  def normalize(self, x):
    norm_x = (x - self.mean) / self.std
    norm_x = np.clip(norm_x, -self.clip, self.clip)
    return norm_x

  def unnormalize(self, norm_x):
    x = norm_x * self.std + self.mean
    return x

  def calc_std(self, mean, mean_sq):
    var = mean_sq - np.square(mean)
    # some time floating point errors can lead to small negative numbers
    var = np.maximum(var, 0)
    std = np.sqrt(var)
    std = np.maximum(std, self.eps)
    return std

  def calc_mean_sq(self, mean, std):
    return np.square(std) + np.square(self.mean)

  def check_synced(self):
    synced = True
    if MPIUtil.is_root_proc():
      vars = np.concatenate([self.mean, self.mean_sq])
      MPIUtil.bcast(vars)
    else:
      vars_local = np.concatenate([self.mean, self.mean_sq])
      vars_root = np.empty_like(vars_local)
      MPIUtil.bcast(vars_root)
      synced = (vars_local == vars_root).all()
    return synced

  def _build_groups(self, groups_ids):
    groups = []
    if groups_ids is None:
      curr_id = self.NORM_GROUP_SINGLE
      curr_list = np.arange(self.get_size()).astype(np.int32)
      groups.append(self.Group(curr_id, curr_list))
    else:
      ids = np.unique(groups_ids)
      for id in ids:
        curr_list = np.nonzero(groups_ids == id)[0].astype(np.int32)
        groups.append(self.Group(id, curr_list))

    return groups

  def _process_group_data(self, new_data, old_data):
    proc_data = new_data.copy()
    for group in self.groups:
      if group.id == self.NORM_GROUP_NONE:
        proc_data[group.indices] = old_data[group.indices]
      elif group.id != self.NORM_GROUP_SINGLE:
        avg = np.mean(new_data[group.indices])
        proc_data[group.indices] = avg
    return proc_data



# Lint as: python3
"""Add the new laikago robot."""
import gin

from pybullet_envs.minitaur.robots import laikago_constants
from pybullet_envs.minitaur.robots import quadruped_base
from pybullet_envs.minitaur.robots import robot_urdf_loader


@gin.configurable
class Laikago(quadruped_base.QuadrupedBase):
  """The Laikago class that simulates the quadruped from Unitree."""

  def _pre_load(self):
    """Import the Laikago specific constants.
    """
    self._urdf_loader = robot_urdf_loader.RobotUrdfLoader(
        pybullet_client=self._pybullet_client,
        urdf_path=laikago_constants.URDF_PATH,
        enable_self_collision=True,
        init_base_position=laikago_constants.INIT_POSITION,
        init_base_orientation_quaternion=laikago_constants.INIT_ORIENTATION,
        init_joint_angles=laikago_constants.INIT_JOINT_ANGLES,
        joint_offsets=laikago_constants.JOINT_OFFSETS,
        joint_directions=laikago_constants.JOINT_DIRECTIONS,
        motor_names=laikago_constants.MOTOR_NAMES,
        end_effector_names=laikago_constants.END_EFFECTOR_NAMES,
        user_group=laikago_constants.MOTOR_GROUP,
    )

  @classmethod
  def get_constants(cls):
    del cls
    return laikago_constants



import pybullet as p
import math
import numpy as np
import pybullet_data

p.connect(p.GUI)
p.setAdditionalSearchPath(pybullet_data.getDataPath())
plane = p.loadURDF("plane100.urdf")
cube = p.loadURDF("cube.urdf", [0, 0, 1])


def getRayFromTo(mouseX, mouseY):
  width, height, viewMat, projMat, cameraUp, camForward, horizon, vertical, _, _, dist, camTarget = p.getDebugVisualizerCamera(
  )
  camPos = [
      camTarget[0] - dist * camForward[0], camTarget[1] - dist * camForward[1],
      camTarget[2] - dist * camForward[2]
  ]
  farPlane = 10000
  rayForward = [(camTarget[0] - camPos[0]), (camTarget[1] - camPos[1]), (camTarget[2] - camPos[2])]
  lenFwd = math.sqrt(rayForward[0] * rayForward[0] + rayForward[1] * rayForward[1] +
                     rayForward[2] * rayForward[2])
  invLen = farPlane * 1. / lenFwd
  rayForward = [invLen * rayForward[0], invLen * rayForward[1], invLen * rayForward[2]]
  rayFrom = camPos
  oneOverWidth = float(1) / float(width)
  oneOverHeight = float(1) / float(height)

  dHor = [horizon[0] * oneOverWidth, horizon[1] * oneOverWidth, horizon[2] * oneOverWidth]
  dVer = [vertical[0] * oneOverHeight, vertical[1] * oneOverHeight, vertical[2] * oneOverHeight]
  rayToCenter = [
      rayFrom[0] + rayForward[0], rayFrom[1] + rayForward[1], rayFrom[2] + rayForward[2]
  ]
  ortho = [
      -0.5 * horizon[0] + 0.5 * vertical[0] + float(mouseX) * dHor[0] - float(mouseY) * dVer[0],
      -0.5 * horizon[1] + 0.5 * vertical[1] + float(mouseX) * dHor[1] - float(mouseY) * dVer[1],
      -0.5 * horizon[2] + 0.5 * vertical[2] + float(mouseX) * dHor[2] - float(mouseY) * dVer[2]
  ]

  rayTo = [
      rayFrom[0] + rayForward[0] + ortho[0], rayFrom[1] + rayForward[1] + ortho[1],
      rayFrom[2] + rayForward[2] + ortho[2]
  ]
  lenOrtho = math.sqrt(ortho[0] * ortho[0] + ortho[1] * ortho[1] + ortho[2] * ortho[2])
  alpha = math.atan(lenOrtho / farPlane)
  return rayFrom, rayTo, alpha


width, height, viewMat, projMat, cameraUp, camForward, horizon, vertical, _, _, dist, camTarget = p.getDebugVisualizerCamera(
)
camPos = [
    camTarget[0] - dist * camForward[0], camTarget[1] - dist * camForward[1],
    camTarget[2] - dist * camForward[2]
]
farPlane = 10000
rayForward = [(camTarget[0] - camPos[0]), (camTarget[1] - camPos[1]), (camTarget[2] - camPos[2])]
lenFwd = math.sqrt(rayForward[0] * rayForward[0] + rayForward[1] * rayForward[1] +
                   rayForward[2] * rayForward[2])
oneOverWidth = float(1) / float(width)
oneOverHeight = float(1) / float(height)
dHor = [horizon[0] * oneOverWidth, horizon[1] * oneOverWidth, horizon[2] * oneOverWidth]
dVer = [vertical[0] * oneOverHeight, vertical[1] * oneOverHeight, vertical[2] * oneOverHeight]

lendHor = math.sqrt(dHor[0] * dHor[0] + dHor[1] * dHor[1] + dHor[2] * dHor[2])
lendVer = math.sqrt(dVer[0] * dVer[0] + dVer[1] * dVer[1] + dVer[2] * dVer[2])

cornersX = [0, width, width, 0]
cornersY = [0, 0, height, height]
corners3D = []

imgW = int(width / 4)
imgH = int(height / 4)

img = p.getCameraImage(imgW, imgH, renderer=p.ER_BULLET_HARDWARE_OPENGL)
rgbBuffer = img[2]
depthBuffer = img[3]
print("rgbBuffer.shape=", rgbBuffer.shape)
print("depthBuffer.shape=", depthBuffer.shape)

#disable rendering temporary makes adding objects faster
p.configureDebugVisualizer(p.COV_ENABLE_RENDERING, 0)
p.configureDebugVisualizer(p.COV_ENABLE_GUI, 0)
p.configureDebugVisualizer(p.COV_ENABLE_TINY_RENDERER, 0)
visualShapeId = p.createVisualShape(shapeType=p.GEOM_SPHERE, rgbaColor=[1, 1, 1, 1], radius=0.03)
collisionShapeId = -1  #p.createCollisionShape(shapeType=p.GEOM_MESH, fileName="duck_vhacd.obj", collisionFramePosition=shift,meshScale=meshScale)

for i in range(4):
  w = cornersX[i]
  h = cornersY[i]
  rayFrom, rayTo, _ = getRayFromTo(w, h)
  rf = np.array(rayFrom)
  rt = np.array(rayTo)
  vec = rt - rf
  l = np.sqrt(np.dot(vec, vec))
  newTo = (0.01 / l) * vec + rf
  #print("len vec=",np.sqrt(np.dot(vec,vec)))

  p.addUserDebugLine(rayFrom, newTo, [1, 0, 0])
  corners3D.append(newTo)
count = 0

stepX = 5
stepY = 5
for w in range(0, imgW, stepX):
  for h in range(0, imgH, stepY):
    count += 1
    if ((count % 100) == 0):
      print(count, "out of ", imgW * imgH / (stepX * stepY))
    rayFrom, rayTo, alpha = getRayFromTo(w * (width / imgW), h * (height / imgH))
    rf = np.array(rayFrom)
    rt = np.array(rayTo)
    vec = rt - rf
    l = np.sqrt(np.dot(vec, vec))
    depthImg = float(depthBuffer[h, w])
    far = 1000.
    near = 0.01
    depth = far * near / (far - (far - near) * depthImg)
    depth /= math.cos(alpha)
    newTo = (depth / l) * vec + rf
    #p.addUserDebugLine(rayFrom, newTo, [1, 0, 0])
    mb = p.createMultiBody(baseMass=0,
                           baseCollisionShapeIndex=collisionShapeId,
                           baseVisualShapeIndex=visualShapeId,
                           basePosition=newTo,
                           useMaximalCoordinates=True)
    color = rgbBuffer[h, w]
    color = [color[0] / 255., color[1] / 255., color[2] / 255., 1]
    p.changeVisualShape(mb, -1, rgbaColor=color)
p.addUserDebugLine(corners3D[0], corners3D[1], [1, 0, 0])
p.addUserDebugLine(corners3D[1], corners3D[2], [1, 0, 0])
p.addUserDebugLine(corners3D[2], corners3D[3], [1, 0, 0])
p.addUserDebugLine(corners3D[3], corners3D[0], [1, 0, 0])
p.configureDebugVisualizer(p.COV_ENABLE_RENDERING, 1)
print("ready\n")
#p.removeBody(plane)
#p.removeBody(cube)
while (1):
  p.setGravity(0, 0, -10)



# Copyright 2017 The TensorFlow Agents Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for environment wrappers."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import functools

import tf.compat.v1 as tf

from agents import tools


class ExternalProcessTest(tf.test.TestCase):

  def test_close_no_hang_after_init(self):
    constructor = functools.partial(tools.MockEnvironment,
                                    observ_shape=(2, 3),
                                    action_shape=(2,),
                                    min_duration=2,
                                    max_duration=2)
    env = tools.wrappers.ExternalProcess(constructor)
    env.close()

  def test_close_no_hang_after_step(self):
    constructor = functools.partial(tools.MockEnvironment,
                                    observ_shape=(2, 3),
                                    action_shape=(2,),
                                    min_duration=5,
                                    max_duration=5)
    env = tools.wrappers.ExternalProcess(constructor)
    env.reset()
    env.step(env.action_space.sample())
    env.step(env.action_space.sample())
    env.close()

  def test_reraise_exception_in_init(self):
    constructor = MockEnvironmentCrashInInit
    env = tools.wrappers.ExternalProcess(constructor)
    with self.assertRaises(Exception):
      env.step(env.action_space.sample())

  def test_reraise_exception_in_step(self):
    constructor = functools.partial(MockEnvironmentCrashInStep, crash_at_step=3)
    env = tools.wrappers.ExternalProcess(constructor)
    env.reset()
    env.step(env.action_space.sample())
    env.step(env.action_space.sample())
    with self.assertRaises(Exception):
      env.step(env.action_space.sample())


class MockEnvironmentCrashInInit(object):
  """Raise an error when instantiated."""

  def __init__(self, *unused_args, **unused_kwargs):
    raise RuntimeError()


class MockEnvironmentCrashInStep(tools.MockEnvironment):
  """Raise an error after specified number of steps in an episode."""

  def __init__(self, crash_at_step):
    super(MockEnvironmentCrashInStep, self).__init__(observ_shape=(2, 3),
                                                     action_shape=(2,),
                                                     min_duration=crash_at_step + 1,
                                                     max_duration=crash_at_step + 1)
    self._crash_at_step = crash_at_step

  def step(self, *args, **kwargs):
    transition = super(MockEnvironmentCrashInStep, self).step(*args, **kwargs)
    if self.steps[-1] == self._crash_at_step:
      raise RuntimeError()
    return transition


if __name__ == '__main__':
  tf.test.main()



import pybullet as p
import time


class UrdfInertial(object):

  def __init__(self):
    self.mass = 1
    self.inertia_xxyyzz = [7, 8, 9]
    self.origin_rpy = [1, 2, 3]
    self.origin_xyz = [4, 5, 6]


class UrdfContact(object):

  def __init__(self):
    self.lateral_friction = 1
    self.rolling_friction = 0
    self.spinning_friction = 0


class UrdfLink(object):

  def __init__(self):
    self.link_name = "dummy"
    self.urdf_inertial = UrdfInertial()
    self.urdf_visual_shapes = []
    self.urdf_collision_shapes = []


class UrdfVisual(object):

  def __init__(self):
    self.origin_rpy = [1, 2, 3]
    self.origin_xyz = [4, 5, 6]
    self.geom_type = p.GEOM_BOX
    self.geom_radius = 1
    self.geom_extents = [7, 8, 9]
    self.geom_length = 10
    self.geom_meshfilename = "meshfile"
    self.geom_meshscale = [1, 1, 1]
    self.material_rgba = [1, 0, 0, 1]
    self.material_name = ""


class UrdfCollision(object):

  def __init__(self):
    self.origin_rpy = [1, 2, 3]
    self.origin_xyz = [4, 5, 6]
    self.geom_type = p.GEOM_BOX
    self.geom_radius = 1
    self.geom_length = 2
    self.geom_extents = [7, 8, 9]
    self.geom_meshfilename = "meshfile"
    self.geom_meshscale = [1, 1, 1]


class UrdfJoint(object):

  def __init__(self):
    self.link = UrdfLink()
    self.joint_name = "joint_dummy"
    self.joint_type = p.JOINT_REVOLUTE
    self.joint_lower_limit = 0
    self.joint_upper_limit = -1
    self.parent_name = "parentName"
    self.child_name = "childName"
    self.joint_origin_xyz = [1, 2, 3]
    self.joint_origin_rpy = [1, 2, 3]
    self.joint_axis_xyz = [1, 2, 3]


class UrdfEditor(object):

  def __init__(self):
    self.initialize()

  def initialize(self):
    self.urdfLinks = []
    self.urdfJoints = []
    self.robotName = ""
    self.linkNameToIndex = {}
    self.jointTypeToName={p.JOINT_FIXED: "JOINT_FIXED" ,\
               p.JOINT_REVOLUTE: "JOINT_REVOLUTE",\
               p.JOINT_PRISMATIC: "JOINT_PRISMATIC" }

  def convertLinkFromMultiBody(self, bodyUid, linkIndex, urdfLink, physicsClientId):
    dyn = p.getDynamicsInfo(bodyUid, linkIndex, physicsClientId=physicsClientId)
    urdfLink.urdf_inertial.mass = dyn[0]
    urdfLink.urdf_inertial.inertia_xxyyzz = dyn[2]
    #todo
    urdfLink.urdf_inertial.origin_xyz = dyn[3]
    rpy = p.getEulerFromQuaternion(dyn[4])
    urdfLink.urdf_inertial.origin_rpy = rpy

    visualShapes = p.getVisualShapeData(bodyUid, physicsClientId=physicsClientId)
    matIndex = 0
    for v in visualShapes:
      if (v[1] == linkIndex):
        urdfVisual = UrdfVisual()
        urdfVisual.geom_type = v[2]
        if (v[2] == p.GEOM_BOX):
          urdfVisual.geom_extents = v[3]
        if (v[2] == p.GEOM_SPHERE):
          urdfVisual.geom_radius = v[3][0]
        if (v[2] == p.GEOM_MESH):
          urdfVisual.geom_meshfilename = v[4].decode("utf-8")
          urdfVisual.geom_meshscale = v[3]
        if (v[2] == p.GEOM_CYLINDER):
          urdfVisual.geom_length = v[3][0]
          urdfVisual.geom_radius = v[3][1]
        if (v[2] == p.GEOM_CAPSULE):
          urdfVisual.geom_length = v[3][0]
          urdfVisual.geom_radius = v[3][1]
        urdfVisual.origin_xyz = v[5]
        urdfVisual.origin_rpy = p.getEulerFromQuaternion(v[6])
        urdfVisual.material_rgba = v[7]
        name = 'mat_{}_{}'.format(linkIndex, matIndex)
        urdfVisual.material_name = name
        urdfLink.urdf_visual_shapes.append(urdfVisual)
        matIndex = matIndex + 1

    collisionShapes = p.getCollisionShapeData(bodyUid, linkIndex, physicsClientId=physicsClientId)
    for v in collisionShapes:
      urdfCollision = UrdfCollision()
      urdfCollision.geom_type = v[2]
      if (v[2] == p.GEOM_BOX):
        urdfCollision.geom_extents = v[3]
      if (v[2] == p.GEOM_SPHERE):
        urdfCollision.geom_radius = v[3][0]
      if (v[2] == p.GEOM_MESH):
        urdfCollision.geom_meshfilename = v[4].decode("utf-8")
        urdfCollision.geom_meshscale = v[3]
      if (v[2] == p.GEOM_CYLINDER):
        urdfCollision.geom_length = v[3][0]
        urdfCollision.geom_radius = v[3][1]
      if (v[2] == p.GEOM_CAPSULE):
        urdfCollision.geom_length = v[3][0]
        urdfCollision.geom_radius = v[3][1]
      pos,orn = p.multiplyTransforms(dyn[3],dyn[4],\
       v[5], v[6])
      urdfCollision.origin_xyz = pos
      urdfCollision.origin_rpy = p.getEulerFromQuaternion(orn)
      urdfLink.urdf_collision_shapes.append(urdfCollision)

  def initializeFromBulletBody(self, bodyUid, physicsClientId):
    self.initialize()

    #always create a base link
    baseLink = UrdfLink()
    baseLinkIndex = -1
    self.convertLinkFromMultiBody(bodyUid, baseLinkIndex, baseLink, physicsClientId)
    baseLink.link_name = p.getBodyInfo(bodyUid, physicsClientId=physicsClientId)[0].decode("utf-8")
    self.robotName = p.getBodyInfo(bodyUid, physicsClientId=physicsClientId)[1].decode("utf-8")
    self.linkNameToIndex[baseLink.link_name] = len(self.urdfLinks)
    self.urdfLinks.append(baseLink)

    #optionally create child links and joints
    for j in range(p.getNumJoints(bodyUid, physicsClientId=physicsClientId)):
      jointInfo = p.getJointInfo(bodyUid, j, physicsClientId=physicsClientId)
      urdfLink = UrdfLink()
      self.convertLinkFromMultiBody(bodyUid, j, urdfLink, physicsClientId)
      urdfLink.link_name = jointInfo[12].decode("utf-8")
      self.linkNameToIndex[urdfLink.link_name] = len(self.urdfLinks)
      self.urdfLinks.append(urdfLink)

      urdfJoint = UrdfJoint()
      urdfJoint.link = urdfLink
      urdfJoint.joint_name = jointInfo[1].decode("utf-8")
      urdfJoint.joint_type = jointInfo[2]
      urdfJoint.joint_lower_limit = jointInfo[8]
      urdfJoint.joint_upper_limit = jointInfo[9]
      urdfJoint.joint_axis_xyz = jointInfo[13]
      orgParentIndex = jointInfo[16]
      if (orgParentIndex < 0):
        urdfJoint.parent_name = baseLink.link_name
      else:
        parentJointInfo = p.getJointInfo(bodyUid, orgParentIndex, physicsClientId=physicsClientId)
        urdfJoint.parent_name = parentJointInfo[12].decode("utf-8")
      urdfJoint.child_name = urdfLink.link_name

      #todo, compensate for inertia/link frame offset

      dynChild = p.getDynamicsInfo(bodyUid, j, physicsClientId=physicsClientId)
      childInertiaPos = dynChild[3]
      childInertiaOrn = dynChild[4]
      parentCom2JointPos = jointInfo[14]
      parentCom2JointOrn = jointInfo[15]
      tmpPos, tmpOrn = p.multiplyTransforms(childInertiaPos, childInertiaOrn, parentCom2JointPos,
                                            parentCom2JointOrn)
      tmpPosInv, tmpOrnInv = p.invertTransform(tmpPos, tmpOrn)
      dynParent = p.getDynamicsInfo(bodyUid, orgParentIndex, physicsClientId=physicsClientId)
      parentInertiaPos = dynParent[3]
      parentInertiaOrn = dynParent[4]

      pos, orn = p.multiplyTransforms(parentInertiaPos, parentInertiaOrn, tmpPosInv, tmpOrnInv)
      pos, orn_unused = p.multiplyTransforms(parentInertiaPos, parentInertiaOrn,
                                             parentCom2JointPos, [0, 0, 0, 1])

      urdfJoint.joint_origin_xyz = pos
      urdfJoint.joint_origin_rpy = p.getEulerFromQuaternion(orn)

      self.urdfJoints.append(urdfJoint)

  def writeInertial(self, file, urdfInertial, precision=5):
    file.write("\t\t<inertial>\n")
    str = '\t\t\t<origin rpy=\"{:.{prec}f} {:.{prec}f} {:.{prec}f}\" xyz=\"{:.{prec}f} {:.{prec}f} {:.{prec}f}\"/>\n'.format(\
    urdfInertial.origin_rpy[0],urdfInertial.origin_rpy[1],urdfInertial.origin_rpy[2],\
    urdfInertial.origin_xyz[0],urdfInertial.origin_xyz[1],urdfInertial.origin_xyz[2], prec=precision)
    file.write(str)
    str = '\t\t\t<mass value=\"{:.{prec}f}\"/>\n'.format(urdfInertial.mass, prec=precision)
    file.write(str)
    str = '\t\t\t<inertia ixx=\"{:.{prec}f}\" ixy=\"0\" ixz=\"0\" iyy=\"{:.{prec}f}\" iyz=\"0\" izz=\"{:.{prec}f}\"/>\n'.format(\
    urdfInertial.inertia_xxyyzz[0],\
    urdfInertial.inertia_xxyyzz[1],\
    urdfInertial.inertia_xxyyzz[2],prec=precision)
    file.write(str)
    file.write("\t\t</inertial>\n")

  def writeVisualShape(self, file, urdfVisual, precision=5):
    #we don't support loading capsule types from visuals, so auto-convert from collision shape
    if urdfVisual.geom_type == p.GEOM_CAPSULE:
      return

    file.write("\t\t<visual>\n")
    str = '\t\t\t<origin rpy="{:.{prec}f} {:.{prec}f} {:.{prec}f}" xyz="{:.{prec}f} {:.{prec}f} {:.{prec}f}"/>\n'.format(\
     urdfVisual.origin_rpy[0],urdfVisual.origin_rpy[1],urdfVisual.origin_rpy[2],
     urdfVisual.origin_xyz[0],urdfVisual.origin_xyz[1],urdfVisual.origin_xyz[2], prec=precision)
    file.write(str)
    file.write("\t\t\t<geometry>\n")
    if urdfVisual.geom_type == p.GEOM_BOX:
      str = '\t\t\t\t<box size=\"{:.{prec}f} {:.{prec}f} {:.{prec}f}\"/>\n'.format(urdfVisual.geom_extents[0],\
       urdfVisual.geom_extents[1],urdfVisual.geom_extents[2], prec=precision)
      file.write(str)
    if urdfVisual.geom_type == p.GEOM_SPHERE:
      str = '\t\t\t\t<sphere radius=\"{:.{prec}f}\"/>\n'.format(urdfVisual.geom_radius,\
       prec=precision)
      file.write(str)
    if urdfVisual.geom_type == p.GEOM_MESH:

      str = '\t\t\t\t<mesh filename=\"{}\" scale=\"{:.{prec}f} {:.{prec}f} {:.{prec}f}\"/>\n'.format(
      urdfVisual.geom_meshfilename,urdfVisual.geom_meshscale[0],urdfVisual.geom_meshscale[1],urdfVisual.geom_meshscale[2],\
       prec=precision)
      file.write(str)
    if urdfVisual.geom_type == p.GEOM_CYLINDER:
      str = '\t\t\t\t<cylinder length=\"{:.{prec}f}\" radius=\"{:.{prec}f}\"/>\n'.format(\
       urdfVisual.geom_length, urdfVisual.geom_radius, prec=precision)
      file.write(str)
    if urdfVisual.geom_type == p.GEOM_CAPSULE:
      str = '\t\t\t\t<capsule length=\"{:.{prec}f}\" radius=\"{:.{prec}f}\"/>\n'.format(\
       urdfVisual.geom_length, urdfVisual.geom_radius, prec=precision)
      file.write(str)

    file.write("\t\t\t</geometry>\n")
    str = '\t\t\t<material name=\"{}\">\n'.format(urdfVisual.material_name)
    file.write(str)
    str = '\t\t\t\t<color rgba="{:.{prec}f} {:.{prec}f} {:.{prec}f} {:.{prec}f}" />\n'.format(urdfVisual.material_rgba[0],\
     urdfVisual.material_rgba[1],urdfVisual.material_rgba[2],urdfVisual.material_rgba[3],prec=precision)
    file.write(str)
    file.write("\t\t\t</material>\n")
    file.write("\t\t</visual>\n")

  def writeCollisionShape(self, file, urdfCollision, precision=5):
    file.write("\t\t<collision>\n")
    str = '\t\t\t<origin rpy="{:.{prec}f} {:.{prec}f} {:.{prec}f}" xyz="{:.{prec}f} {:.{prec}f} {:.{prec}f}"/>\n'.format(\
     urdfCollision.origin_rpy[0],urdfCollision.origin_rpy[1],urdfCollision.origin_rpy[2],
     urdfCollision.origin_xyz[0],urdfCollision.origin_xyz[1],urdfCollision.origin_xyz[2], prec=precision)
    file.write(str)
    file.write("\t\t\t<geometry>\n")
    if urdfCollision.geom_type == p.GEOM_BOX:
      str = '\t\t\t\t<box size=\"{:.{prec}f} {:.{prec}f} {:.{prec}f}\"/>\n'.format(urdfCollision.geom_extents[0],\
       urdfCollision.geom_extents[1],urdfCollision.geom_extents[2], prec=precision)
      file.write(str)
    if urdfCollision.geom_type == p.GEOM_SPHERE:
      str = '\t\t\t\t<sphere radius=\"{:.{prec}f}\"/>\n'.format(urdfCollision.geom_radius,\
       prec=precision)
      file.write(str)
    if urdfCollision.geom_type == p.GEOM_MESH:
      str = '\t\t\t\t<mesh filename=\"{}\" scale=\"{:.{prec}f} {:.{prec}f} {:.{prec}f}\"/>\n'.format(urdfCollision.geom_meshfilename,\
       urdfCollision.geom_meshscale[0],urdfCollision.geom_meshscale[1],urdfCollision.geom_meshscale[2],prec=precision)
      file.write(str)
    if urdfCollision.geom_type == p.GEOM_CYLINDER:
      str = '\t\t\t\t<cylinder length=\"{:.{prec}f}\" radius=\"{:.{prec}f}\"/>\n'.format(\
       urdfCollision.geom_length, urdfCollision.geom_radius, prec=precision)
      file.write(str)
    if urdfCollision.geom_type == p.GEOM_CAPSULE:
      str = '\t\t\t\t<capsule length=\"{:.{prec}f}\" radius=\"{:.{prec}f}\"/>\n'.format(\
       urdfCollision.geom_length, urdfCollision.geom_radius, prec=precision)
      file.write(str)
    file.write("\t\t\t</geometry>\n")
    file.write("\t\t</collision>\n")

  def writeLink(self, file, urdfLink, saveVisuals):
    file.write("\t<link name=\"")
    file.write(urdfLink.link_name)
    file.write("\">\n")

    self.writeInertial(file, urdfLink.urdf_inertial)
    hasCapsules = False
    for v in urdfLink.urdf_visual_shapes:
      if (v.geom_type == p.GEOM_CAPSULE):
        hasCapsules = True
    if (saveVisuals and not hasCapsules):
      for v in urdfLink.urdf_visual_shapes:
        self.writeVisualShape(file, v)
    for c in urdfLink.urdf_collision_shapes:
      self.writeCollisionShape(file, c)
    file.write("\t</link>\n")

  def writeJoint(self, file, urdfJoint, precision=5):
    jointTypeStr = "invalid"
    if urdfJoint.joint_type == p.JOINT_REVOLUTE:
      if urdfJoint.joint_upper_limit < urdfJoint.joint_lower_limit:
        jointTypeStr = "continuous"
      else:
        jointTypeStr = "revolute"
    if urdfJoint.joint_type == p.JOINT_FIXED:
      jointTypeStr = "fixed"
    if urdfJoint.joint_type == p.JOINT_PRISMATIC:
      jointTypeStr = "prismatic"
    str = '\t<joint name=\"{}\" type=\"{}\">\n'.format(urdfJoint.joint_name, jointTypeStr)
    file.write(str)
    str = '\t\t<parent link=\"{}\"/>\n'.format(urdfJoint.parent_name)
    file.write(str)
    str = '\t\t<child link=\"{}\"/>\n'.format(urdfJoint.child_name)
    file.write(str)

    if urdfJoint.joint_type == p.JOINT_PRISMATIC:
      #todo: handle limits
      lowerLimit = -0.5
      upperLimit = 0.5
      str = '<limit effort="1000.0" lower="{:.{prec}f}" upper="{:.{prec}f}" velocity="0.5"/>'.format(
          lowerLimit, upperLimit, prec=precision)
      file.write(str)

    file.write("\t\t<dynamics damping=\"1.0\" friction=\"0.0001\"/>\n")
    str = '\t\t<origin xyz=\"{:.{prec}f} {:.{prec}f} {:.{prec}f}\"/>\n'.format(urdfJoint.joint_origin_xyz[0],\
     urdfJoint.joint_origin_xyz[1],urdfJoint.joint_origin_xyz[2], prec=precision)
    str = '\t\t<origin rpy=\"{:.{prec}f} {:.{prec}f} {:.{prec}f}\" xyz=\"{:.{prec}f} {:.{prec}f} {:.{prec}f}\"/>\n'.format(urdfJoint.joint_origin_rpy[0],\
     urdfJoint.joint_origin_rpy[1],urdfJoint.joint_origin_rpy[2],\
     urdfJoint.joint_origin_xyz[0],\
     urdfJoint.joint_origin_xyz[1],urdfJoint.joint_origin_xyz[2], prec=precision)

    file.write(str)
    str = '\t\t<axis xyz=\"{:.{prec}f} {:.{prec}f} {:.{prec}f}\"/>\n'.format(urdfJoint.joint_axis_xyz[0],\
     urdfJoint.joint_axis_xyz[1],urdfJoint.joint_axis_xyz[2], prec=precision)
    file.write(str)
    file.write("\t</joint>\n")

  def saveUrdf(self, fileName, saveVisuals=True):
    file = open(fileName, "w")
    file.write("<?xml version=\"0.0\" ?>\n")
    file.write("<robot name=\"")
    file.write(self.robotName)
    file.write("\">\n")

    for link in self.urdfLinks:
      self.writeLink(file, link, saveVisuals)

    for joint in self.urdfJoints:
      self.writeJoint(file, joint)

    file.write("</robot>\n")
    file.close()


#def addLink(...)

  def joinUrdf(self,
               childEditor,
               parentLinkIndex=0,
               jointPivotXYZInParent=[0, 0, 0],
               jointPivotRPYInParent=[0, 0, 0],
               jointPivotXYZInChild=[0, 0, 0],
               jointPivotRPYInChild=[0, 0, 0],
               parentPhysicsClientId=0,
               childPhysicsClientId=0):

    childLinkIndex = len(self.urdfLinks)
    insertJointIndex = len(self.urdfJoints)

    #combine all links, and add a joint

    for link in childEditor.urdfLinks:
      self.linkNameToIndex[link.link_name] = len(self.urdfLinks)
      self.urdfLinks.append(link)
    for joint in childEditor.urdfJoints:
      self.urdfJoints.append(joint)
    #add a new joint between a particular

    jointPivotQuatInChild = p.getQuaternionFromEuler(jointPivotRPYInChild)
    invJointPivotXYZInChild, invJointPivotQuatInChild = p.invertTransform(
        jointPivotXYZInChild, jointPivotQuatInChild)

    #apply this invJointPivot***InChild to all inertial, visual and collision element in the child link
    #inertial
    pos, orn = p.multiplyTransforms(self.urdfLinks[childLinkIndex].urdf_inertial.origin_xyz,
                                    p.getQuaternionFromEuler(
                                        self.urdfLinks[childLinkIndex].urdf_inertial.origin_rpy),
                                    invJointPivotXYZInChild,
                                    invJointPivotQuatInChild,
                                    physicsClientId=parentPhysicsClientId)
    self.urdfLinks[childLinkIndex].urdf_inertial.origin_xyz = pos
    self.urdfLinks[childLinkIndex].urdf_inertial.origin_rpy = p.getEulerFromQuaternion(orn)
    #all visual
    for v in self.urdfLinks[childLinkIndex].urdf_visual_shapes:
      pos, orn = p.multiplyTransforms(v.origin_xyz, p.getQuaternionFromEuler(v.origin_rpy),
                                      invJointPivotXYZInChild, invJointPivotQuatInChild)
      v.origin_xyz = pos
      v.origin_rpy = p.getEulerFromQuaternion(orn)
    #all collision
    for c in self.urdfLinks[childLinkIndex].urdf_collision_shapes:
      pos, orn = p.multiplyTransforms(c.origin_xyz, p.getQuaternionFromEuler(c.origin_rpy),
                                      invJointPivotXYZInChild, invJointPivotQuatInChild)
      c.origin_xyz = pos
      c.origin_rpy = p.getEulerFromQuaternion(orn)

    childLink = self.urdfLinks[childLinkIndex]
    parentLink = self.urdfLinks[parentLinkIndex]

    joint = UrdfJoint()
    joint.link = childLink
    joint.joint_name = "joint_dummy1"
    joint.joint_type = p.JOINT_REVOLUTE
    joint.joint_lower_limit = 0
    joint.joint_upper_limit = -1
    joint.parent_name = parentLink.link_name
    joint.child_name = childLink.link_name
    joint.joint_origin_xyz = jointPivotXYZInParent
    joint.joint_origin_rpy = jointPivotRPYInParent
    joint.joint_axis_xyz = [0, 0, 1]

    #the following commented line would crash PyBullet, it messes up the joint indexing/ordering
    #self.urdfJoints.append(joint)

    #so make sure to insert the joint in the right place, to links/joints match
    self.urdfJoints.insert(insertJointIndex, joint)
    return joint

  def createMultiBody(self,
                      basePosition=[0, 0, 0],
                      baseOrientation=[0, 0, 0, 1],
                      physicsClientId=0):
    #assume link[0] is base
    if (len(self.urdfLinks) == 0):
      return -1

    #for i in range (len(self.urdfLinks)):
    #	print("link", i, "=",self.urdfLinks[i].link_name)

    base = self.urdfLinks[0]

    #v.tmp_collision_shape_ids=[]
    baseMass = base.urdf_inertial.mass
    baseCollisionShapeIndex = -1
    baseShapeTypeArray = []
    baseRadiusArray = []
    baseHalfExtentsArray = []
    lengthsArray = []
    fileNameArray = []
    meshScaleArray = []
    basePositionsArray = []
    baseOrientationsArray = []

    for v in base.urdf_collision_shapes:
      shapeType = v.geom_type
      baseShapeTypeArray.append(shapeType)
      baseHalfExtentsArray.append(
          [0.5 * v.geom_extents[0], 0.5 * v.geom_extents[1], 0.5 * v.geom_extents[2]])
      baseRadiusArray.append(v.geom_radius)
      lengthsArray.append(v.geom_length)
      fileNameArray.append(v.geom_meshfilename)
      meshScaleArray.append(v.geom_meshscale)
      basePositionsArray.append(v.origin_xyz)
      orn = p.getQuaternionFromEuler(v.origin_rpy)
      baseOrientationsArray.append(orn)

    if (len(baseShapeTypeArray)):
      #print("fileNameArray=",fileNameArray)
      baseCollisionShapeIndex = p.createCollisionShapeArray(
          shapeTypes=baseShapeTypeArray,
          radii=baseRadiusArray,
          halfExtents=baseHalfExtentsArray,
          lengths=lengthsArray,
          fileNames=fileNameArray,
          meshScales=meshScaleArray,
          collisionFramePositions=basePositionsArray,
          collisionFrameOrientations=baseOrientationsArray,
          physicsClientId=physicsClientId)

    urdfVisuals = base.urdf_visual_shapes

    shapeTypes = [v.geom_type for v in urdfVisuals]
    halfExtents = [[ext * 0.5 for ext in v.geom_extents] for v in urdfVisuals]
    radii = [v.geom_radius for v in urdfVisuals]
    lengths = [v.geom_length for v in urdfVisuals]
    fileNames = [v.geom_meshfilename for v in urdfVisuals]
    meshScales = [v.geom_meshscale for v in urdfVisuals]
    rgbaColors = [v.material_rgba for v in urdfVisuals]
    visualFramePositions = [v.origin_xyz for v in urdfVisuals]
    visualFrameOrientations = [p.getQuaternionFromEuler(v.origin_rpy) for v in urdfVisuals]
    baseVisualShapeIndex = -1

    if (len(shapeTypes)):
      #print("len(shapeTypes)=",len(shapeTypes))
      #print("len(halfExtents)=",len(halfExtents))
      #print("len(radii)=",len(radii))
      #print("len(lengths)=",len(lengths))
      #print("len(fileNames)=",len(fileNames))
      #print("len(meshScales)=",len(meshScales))
      #print("len(rgbaColors)=",len(rgbaColors))
      #print("len(visualFramePositions)=",len(visualFramePositions))
      #print("len(visualFrameOrientations)=",len(visualFrameOrientations))

      baseVisualShapeIndex = p.createVisualShapeArray(
          shapeTypes=shapeTypes,
          halfExtents=halfExtents,
          radii=radii,
          lengths=lengths,
          fileNames=fileNames,
          meshScales=meshScales,
          rgbaColors=rgbaColors,
          visualFramePositions=visualFramePositions,
          visualFrameOrientations=visualFrameOrientations,
          physicsClientId=physicsClientId)

    linkMasses = []
    linkCollisionShapeIndices = []
    linkVisualShapeIndices = []
    linkPositions = []
    linkOrientations = []

    linkInertialFramePositions = []
    linkInertialFrameOrientations = []
    linkParentIndices = []
    linkJointTypes = []
    linkJointAxis = []

    for joint in self.urdfJoints:
      link = joint.link
      linkMass = link.urdf_inertial.mass
      linkCollisionShapeIndex = -1
      linkVisualShapeIndex = -1
      linkPosition = [0, 0, 0]
      linkOrientation = [0, 0, 0]
      linkInertialFramePosition = [0, 0, 0]
      linkInertialFrameOrientation = [0, 0, 0]
      linkParentIndex = self.linkNameToIndex[joint.parent_name]
      linkJointType = joint.joint_type
      linkJointAx = joint.joint_axis_xyz
      linkShapeTypeArray = []
      linkRadiusArray = []
      linkHalfExtentsArray = []
      lengthsArray = []
      fileNameArray = []
      linkMeshScaleArray = []
      linkPositionsArray = []
      linkOrientationsArray = []

      for v in link.urdf_collision_shapes:
        shapeType = v.geom_type
        linkShapeTypeArray.append(shapeType)
        linkHalfExtentsArray.append(
            [0.5 * v.geom_extents[0], 0.5 * v.geom_extents[1], 0.5 * v.geom_extents[2]])
        linkRadiusArray.append(v.geom_radius)
        lengthsArray.append(v.geom_length)
        fileNameArray.append(v.geom_meshfilename)
        linkMeshScaleArray.append(v.geom_meshscale)
        linkPositionsArray.append(v.origin_xyz)
        linkOrientationsArray.append(p.getQuaternionFromEuler(v.origin_rpy))

      if (len(linkShapeTypeArray)):
        linkCollisionShapeIndex = p.createCollisionShapeArray(
            shapeTypes=linkShapeTypeArray,
            radii=linkRadiusArray,
            halfExtents=linkHalfExtentsArray,
            lengths=lengthsArray,
            fileNames=fileNameArray,
            meshScales=linkMeshScaleArray,
            collisionFramePositions=linkPositionsArray,
            collisionFrameOrientations=linkOrientationsArray,
            physicsClientId=physicsClientId)

      urdfVisuals = link.urdf_visual_shapes
      linkVisualShapeIndex = -1
      shapeTypes = [v.geom_type for v in urdfVisuals]
      halfExtents = [[ext * 0.5 for ext in v.geom_extents] for v in urdfVisuals]
      radii = [v.geom_radius for v in urdfVisuals]
      lengths = [v.geom_length for v in urdfVisuals]
      fileNames = [v.geom_meshfilename for v in urdfVisuals]
      meshScales = [v.geom_meshscale for v in urdfVisuals]
      rgbaColors = [v.material_rgba for v in urdfVisuals]
      visualFramePositions = [v.origin_xyz for v in urdfVisuals]
      visualFrameOrientations = [p.getQuaternionFromEuler(v.origin_rpy) for v in urdfVisuals]

      if (len(shapeTypes)):
        print("fileNames=", fileNames)

        linkVisualShapeIndex = p.createVisualShapeArray(
            shapeTypes=shapeTypes,
            halfExtents=halfExtents,
            radii=radii,
            lengths=lengths,
            fileNames=fileNames,
            meshScales=meshScales,
            rgbaColors=rgbaColors,
            visualFramePositions=visualFramePositions,
            visualFrameOrientations=visualFrameOrientations,
            physicsClientId=physicsClientId)

      linkMasses.append(linkMass)
      linkCollisionShapeIndices.append(linkCollisionShapeIndex)
      linkVisualShapeIndices.append(linkVisualShapeIndex)
      linkPositions.append(joint.joint_origin_xyz)
      linkOrientations.append(p.getQuaternionFromEuler(joint.joint_origin_rpy))
      linkInertialFramePositions.append(link.urdf_inertial.origin_xyz)
      linkInertialFrameOrientations.append(p.getQuaternionFromEuler(link.urdf_inertial.origin_rpy))
      linkParentIndices.append(linkParentIndex)
      linkJointTypes.append(joint.joint_type)
      linkJointAxis.append(joint.joint_axis_xyz)
    obUid = p.createMultiBody(baseMass,\
       baseCollisionShapeIndex=baseCollisionShapeIndex,
                                          baseVisualShapeIndex=baseVisualShapeIndex,
       basePosition=basePosition,
       baseOrientation=baseOrientation,
       baseInertialFramePosition=base.urdf_inertial.origin_xyz,
       baseInertialFrameOrientation=p.getQuaternionFromEuler(base.urdf_inertial.origin_rpy),
       linkMasses=linkMasses,
       linkCollisionShapeIndices=linkCollisionShapeIndices,
       linkVisualShapeIndices=linkVisualShapeIndices,
       linkPositions=linkPositions,
       linkOrientations=linkOrientations,
       linkInertialFramePositions=linkInertialFramePositions,
       linkInertialFrameOrientations=linkInertialFrameOrientations,
       linkParentIndices=linkParentIndices,
       linkJointTypes=linkJointTypes,
       linkJointAxis=linkJointAxis,
       physicsClientId=physicsClientId)
    return obUid

  def __del__(self):
    pass



# Copyright 2017 The TensorFlow Agents Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Example configurations using the PPO algorithm."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=unused-variable

from pybullet_envs.minitaur.agents import ppo
from pybullet_envs.minitaur.agents.scripts import networks


def default():
  """Default configuration for PPO."""
  # General
  algorithm = ppo.PPOAlgorithm
  num_agents = 10
  eval_episodes = 25
  use_gpu = False
  # Network
  network = networks.ForwardGaussianPolicy
  weight_summaries = dict(all=r'.*', policy=r'.*/policy/.*', value=r'.*/value/.*')
  policy_layers = 200, 100
  value_layers = 200, 100
  init_mean_factor = 0.05
  init_logstd = -1
  # Optimization
  update_every = 25
  policy_optimizer = 'AdamOptimizer'
  value_optimizer = 'AdamOptimizer'
  update_epochs_policy = 50
  update_epochs_value = 50
  policy_lr = 1e-4
  value_lr = 3e-4
  # Losses
  discount = 0.985
  kl_target = 1e-2
  kl_cutoff_factor = 2
  kl_cutoff_coef = 1000
  kl_init_penalty = 1
  return locals()


def pendulum():
  """Configuration for the pendulum classic control task."""
  locals().update(default())
  # Environment
  env = 'Pendulum-v0'
  max_length = 200
  steps = 1e6  # 1M
  return locals()


def cheetah():
  """Configuration for MuJoCo's half cheetah task."""
  locals().update(default())
  # Environment
  env = 'HalfCheetah-v1'
  max_length = 1000
  steps = 1e7  # 10M
  return locals()


def walker():
  """Configuration for MuJoCo's walker task."""
  locals().update(default())
  # Environment
  env = 'Walker2d-v1'
  max_length = 1000
  steps = 1e7  # 10M
  return locals()


def reacher():
  """Configuration for MuJoCo's reacher task."""
  locals().update(default())
  # Environment
  env = 'Reacher-v1'
  max_length = 1000
  steps = 1e7  # 10M
  return locals()


def hopper():
  """Configuration for MuJoCo's hopper task."""
  locals().update(default())
  # Environment
  env = 'Hopper-v1'
  max_length = 1000
  steps = 2e7  # 20M
  return locals()


def ant():
  """Configuration for MuJoCo's ant task."""
  locals().update(default())
  # Environment
  env = 'Ant-v1'
  max_length = 1000
  steps = 5e7  # 50M
  return locals()


def humanoid():
  """Configuration for MuJoCo's humanoid task."""
  locals().update(default())
  # Environment
  env = 'Humanoid-v1'
  max_length = 1000
  steps = 5e7  # 50M
  return locals()



# Lint as: python3
"""Pybullet simulation of Minitaur robot."""
import math
from typing import Dict, Tuple, Union, Text

from absl import logging
import gin

from pybullet_envs.minitaur.robots import minitaur_constants
from pybullet_envs.minitaur.robots import quadruped_base
from pybullet_envs.minitaur.robots import robot_config
from pybullet_envs.minitaur.robots import robot_urdf_loader


@gin.configurable
class Minitaur(quadruped_base.QuadrupedBase):
  """Minitaur simulation model in pyBullet."""

  def _pre_load(self):
    try:
      use_constrained_base = gin.query_parameter(
          "robot_urdf_loader.RobotUrdfLoader.constrained_base")
    except ValueError:
      use_constrained_base = False
    if use_constrained_base:
      logging.warn(
          "use_constrained_base is currently not compatible with Minitaur's "
          "leg constraints."
      )

    self._urdf_loader = robot_urdf_loader.RobotUrdfLoader(
        pybullet_client=self._pybullet_client,
        enable_self_collision=True,
        urdf_path=minitaur_constants.MINITAUR_URDF_PATH,
        init_base_position=minitaur_constants.INIT_POSITION,
        init_base_orientation_quaternion=minitaur_constants
        .INIT_ORIENTATION_QUAT,
        init_base_orientation_rpy=minitaur_constants.INIT_ORIENTATION_RPY,
        init_joint_angles=minitaur_constants.INIT_JOINT_ANGLES,
        joint_offsets=minitaur_constants.JOINT_OFFSETS,
        joint_directions=minitaur_constants.JOINT_DIRECTIONS,
        motor_names=minitaur_constants.MOTOR_NAMES,
        end_effector_names=minitaur_constants.END_EFFECTOR_NAMES,
        user_group=minitaur_constants.MOTOR_GROUP,
    )

  def _on_load(self):
    """Add hinge constraint for Minitaur's diamond shaped leg after loading."""
    half_pi = math.pi / 2.0
    knee_angle = -2.1834
    for (leg_id, leg_position) in enumerate(minitaur_constants.LEG_ORDER):
      self._pybullet_client.resetJointState(
          self._urdf_loader.robot_id,
          self._joint_id_dict["motor_" + leg_position + "L_joint"],
          self._motor_directions[2 * leg_id] * half_pi,
          targetVelocity=0)
      self._pybullet_client.resetJointState(
          self._urdf_loader.robot_id,
          self._joint_id_dict["knee_" + leg_position + "L_joint"],
          self._motor_directions[2 * leg_id] * knee_angle,
          targetVelocity=0)
      self._pybullet_client.resetJointState(
          self._urdf_loader.robot_id,
          self._joint_id_dict["motor_" + leg_position + "R_joint"],
          self._motor_directions[2 * leg_id + 1] * half_pi,
          targetVelocity=0)
      self._pybullet_client.resetJointState(
          self._urdf_loader.robot_id,
          self._joint_id_dict["knee_" + leg_position + "R_joint"],
          self._motor_directions[2 * leg_id + 1] * knee_angle,
          targetVelocity=0)

      if leg_id < 2:
        self._pybullet_client.createConstraint(
            self._urdf_loader.robot_id,
            self._joint_id_dict["knee_" + leg_position + "R_joint"],
            self._urdf_loader.robot_id,
            self._joint_id_dict["knee_" + leg_position + "L_joint"],
            self._pybullet_client.JOINT_POINT2POINT, [0, 0, 0],
            minitaur_constants.KNEE_CONSTRAINT_POINT_SHORT,
            minitaur_constants.KNEE_CONSTRAINT_POINT_LONG)
      else:
        self._pybullet_client.createConstraint(
            self._urdf_loader.robot_id,
            self._joint_id_dict["knee_" + leg_position + "R_joint"],
            self._urdf_loader.robot_id,
            self._joint_id_dict["knee_" + leg_position + "L_joint"],
            self._pybullet_client.JOINT_POINT2POINT, [0, 0, 0],
            minitaur_constants.KNEE_CONSTRAINT_POINT_LONG,
            minitaur_constants.KNEE_CONSTRAINT_POINT_SHORT)
    self.receive_observation()

  def _reset_joint_angles(self,
                          joint_angles: Union[Tuple[float], Dict[Text,
                                                                 float]] = None,
                          num_reset_steps: int = 100):
    """Resets joint angles of the robot.

    Note that since Minitaur has additional leg constraints on the end
    effectors, directly setting joint angles will lead to constraint violation.
    Instead, we apply motor commands to move the motors to the desired position.

    Args:
      joint_angles: the desired joint angles to reset to.
      num_reset_steps: number of reset steps.
    """
    if joint_angles is None:
      joint_angles = minitaur_constants.INIT_JOINT_ANGLES
    actions = joint_angles
    if isinstance(joint_angles, dict):
      actions = [
          joint_angles[joint_name]
          for joint_name in minitaur_constants.JOINT_NAMES
      ]

    # TODO(b/157786642): since the simulation clock is not stepped here, this
    # reset behaves slightly different compared to the old robot class.
    for _ in range(num_reset_steps):
      self.apply_action(
          actions, motor_control_mode=robot_config.MotorControlMode.POSITION)
      self._pybullet_client.stepSimulation()
      self.receive_observation()



"""The generic safety checking interface.

Defines the generic safety checker class that can detect bad motor states, imu
states, self-collisions, unsafe motor commands, unusual temperature reading,
etc. Safety criterions are provided by the robot class.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import typing
from pybullet_envs.minitaur.robots import robot_config
from pybullet_envs.minitaur.robots.safety import data_types
from pybullet_envs.minitaur.robots.safety import motor_action_validator
from pybullet_envs.minitaur.robots.safety import motor_state_validator
from pybullet_envs.minitaur.robots.safety import utilities

_MOTOR_STATE_BUFFER_SIZE = 500


class SafetyChecker(object):
  """The generic safety checking interface."""

  def __init__(
      self,
      robot: typing.Any,
  ):
    """Initilaizes the class.

    TODO(b/131377892): Implement other state checkings including the
    IMU/temperature/contact force if enabled.

    Args:
      robot: A robot instance such like Minitaur/Laikago/Vision60.
    """
    self._robot = robot

    self._motor_state_validators = []
    self._motor_action_validators = []
    for i in range(robot.num_motors):
      self._motor_state_validators.append(
          motor_state_validator.MotorStateValidator(
              motor_id=i,
              position_bound=robot.safety_config.motor_position_bound[i],
              position_gain_bound=robot.safety_config
              .motor_position_gain_bound[i],
              velocity_bound=robot.safety_config.motor_velocity_bound[i],
              velocity_gain_bound=robot.safety_config
              .motor_velocity_gain_bound[i],
              torque_bound=robot.safety_config.motor_torque_bound[i],
              timestamp_delta_bound=robot.safety_config.timestamp_delta_bound,
              average_abs_velocity_bound=robot.safety_config
              .motor_average_abs_velocity_bound[i],
              average_abs_power_bound=robot.safety_config
              .motor_average_abs_power_bound[i],
              state_buffer_size=_MOTOR_STATE_BUFFER_SIZE,
          ))
      self._motor_action_validators.append(
          motor_action_validator.MotorActionValidator(
              motor_id=i,
              position_bound=robot.safety_config.motor_position_bound[i],
              position_gain_bound=robot.safety_config
              .motor_position_gain_bound[i],
              velocity_bound=robot.safety_config.motor_velocity_bound[i],
              velocity_gain_bound=robot.safety_config
              .motor_velocity_gain_bound[i],
              torque_bound=robot.safety_config.motor_torque_bound[i],
              timestamp_delta_bound=robot.safety_config
              .state_action_timestamp_delta_bound,
              delta_position_bound=robot.safety_config
              .motor_delta_position_bound[i],
              average_abs_delta_position_bound=robot.safety_config
              .motor_average_abs_delta_position_bound[i],
              state_buffer_size=_MOTOR_STATE_BUFFER_SIZE,
          ))

  def check_state(self) -> None:
    """Validates the state of the robot.

    TODO(b/131377892): Implement other state checkings including the
    IMU/temperature/contact force if enabled.

    Raises:
      A safety exception if any state checking (motor/imu/etc) fails.
    """

    for motor_id, state_validator, action_validator in zip(
        range(self._robot.num_motors), self._motor_state_validators,
        self._motor_action_validators):
      motor_state = data_types.MotorState(
          timestamp=self._robot.last_state_time,
          position=self._robot.GetMotorAngles()[motor_id],
          velocity=self._robot.GetMotorVelocities()[motor_id],
          position_gain=self._robot.GetMotorPositionGains()[motor_id],
          velocity_gain=self._robot.GetMotorVelocityGains()[motor_id],
          torque=self._robot.GetMotorTorques()[motor_id],
      )
      state_validator.on_state(motor_state)
      action_validator.on_state(motor_state)

  def check_motor_action(
      self,
      action: typing.Sequence[float],
      control_mode: robot_config.MotorControlMode,
  ) -> None:
    """Validate the action w.r.t to the motor states.

    Args:
      action: The motor commands sent to the robot.
      control_mode: The motor control mode.

    Raises:
      A safety exception if action checking fails.
    """
    motor_action_list = utilities.convert_to_motor_action(
        self._robot, action, control_mode)
    for motor_id, validator in enumerate(self._motor_action_validators):
      validator.on_action(motor_action_list[motor_id], control_mode)



# Copyright 2017 The TensorFlow Agents Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the training loop."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tf.compat.v1 as tf

from google3.robotics.reinforcement_learning.agents import tools


class LoopTest(tf.test.TestCase):

  def test_report_every_step(self):
    step = tf.Variable(0, False, dtype=tf.int32, name='step')
    loop = tools.Loop(None, step)
    loop.add_phase('phase_1', done=True, score=0, summary='', steps=1, report_every=3)
    # Step:   0 1 2 3 4 5 6 7 8
    # Report:     x     x     x
    with self.test_session() as sess:
      sess.run(tf.global_variables_initializer())
      scores = loop.run(sess, saver=None, max_step=9)
      next(scores)
      self.assertEqual(3, sess.run(step))
      next(scores)
      self.assertEqual(6, sess.run(step))
      next(scores)
      self.assertEqual(9, sess.run(step))

  def test_phases_feed(self):
    score = tf.placeholder(tf.float32, [])
    loop = tools.Loop(None)
    loop.add_phase('phase_1',
                   done=True,
                   score=score,
                   summary='',
                   steps=1,
                   report_every=1,
                   log_every=None,
                   checkpoint_every=None,
                   feed={score: 1})
    loop.add_phase('phase_2',
                   done=True,
                   score=score,
                   summary='',
                   steps=3,
                   report_every=1,
                   log_every=None,
                   checkpoint_every=None,
                   feed={score: 2})
    loop.add_phase('phase_3',
                   done=True,
                   score=score,
                   summary='',
                   steps=2,
                   report_every=1,
                   log_every=None,
                   checkpoint_every=None,
                   feed={score: 3})
    with self.test_session() as sess:
      sess.run(tf.global_variables_initializer())
      scores = list(loop.run(sess, saver=None, max_step=15))
    self.assertAllEqual([1, 2, 2, 2, 3, 3, 1, 2, 2, 2, 3, 3, 1, 2, 2], scores)

  def test_average_score_over_phases(self):
    loop = tools.Loop(None)
    loop.add_phase('phase_1', done=True, score=1, summary='', steps=1, report_every=2)
    loop.add_phase('phase_2', done=True, score=2, summary='', steps=2, report_every=5)
    # Score:    1 2 2 1 2 2 1 2 2 1 2 2 1 2 2 1 2
    # Report 1:       x           x           x
    # Report 2:               x             x
    with self.test_session() as sess:
      sess.run(tf.global_variables_initializer())
      scores = list(loop.run(sess, saver=None, max_step=17))
    self.assertAllEqual([1, 2, 1, 2, 1], scores)

  def test_not_done(self):
    step = tf.Variable(0, False, dtype=tf.int32, name='step')
    done = tf.equal((step + 1) % 2, 0)
    score = tf.cast(step, tf.float32)
    loop = tools.Loop(None, step)
    loop.add_phase('phase_1', done, score, summary='', steps=1, report_every=3)
    # Score:  0 1 2 3 4 5 6 7 8
    # Done:     x   x   x   x
    # Report:     x     x     x
    with self.test_session() as sess:
      sess.run(tf.global_variables_initializer())
      scores = list(loop.run(sess, saver=None, max_step=9))
    self.assertAllEqual([1, 4, 7], scores)

  def test_not_done_batch(self):
    step = tf.Variable(0, False, dtype=tf.int32, name='step')
    done = tf.equal([step % 3, step % 4], 0)
    score = tf.cast([step, step**2], tf.float32)
    loop = tools.Loop(None, step)
    loop.add_phase('phase_1', done, score, summary='', steps=1, report_every=8)
    # Step:    0  2  4  6
    # Score 1: 0  2  4  6
    # Done 1:  x        x
    # Score 2: 0  4 16 32
    # Done 2:  x     x
    with self.test_session() as sess:
      sess.run(tf.global_variables_initializer())
      scores = list(loop.run(sess, saver=None, max_step=8))
      self.assertEqual(8, sess.run(step))
    self.assertAllEqual([(0 + 0 + 16 + 6) / 4], scores)


if __name__ == '__main__':
  tf.test.main()



"""State estimator."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import abc


class StateEstimatorBase(object):  # pytype: disable=ignored-metaclass
  """Estimates the unmeasurable state of the robot."""

  __metaclass__ = abc.ABCMeta

  @abc.abstractmethod
  def reset(self, current_time):
    pass

  @abc.abstractmethod
  def update(self, current_time):
    pass



from .scene_abstract import SingleRobotEmptyScene
from .env_bases import MJCFBaseBulletEnv
from robot_pendula import InvertedPendulum, InvertedPendulumSwingup, InvertedDoublePendulum
import gym, gym.spaces, gym.utils, gym.utils.seeding
import numpy as np
import pybullet
import os, sys


class InvertedPendulumBulletEnv(MJCFBaseBulletEnv):

  def __init__(self):
    self.robot = InvertedPendulum()
    MJCFBaseBulletEnv.__init__(self, self.robot)
    self.stateId = -1

  def create_single_player_scene(self, bullet_client):
    return SingleRobotEmptyScene(bullet_client, gravity=9.8, timestep=0.0165, frame_skip=1)

  def reset(self):
    if (self.stateId >= 0):
      #print("InvertedPendulumBulletEnv reset p.restoreState(",self.stateId,")")
      self._p.restoreState(self.stateId)
    r = MJCFBaseBulletEnv.reset(self)
    if (self.stateId < 0):
      self.stateId = self._p.saveState()
      #print("InvertedPendulumBulletEnv reset self.stateId=",self.stateId)
    return r

  def step(self, a):
    self.robot.apply_action(a)
    self.scene.global_step()
    state = self.robot.calc_state()  # sets self.pos_x self.pos_y
    vel_penalty = 0
    if self.robot.swingup:
      reward = np.cos(self.robot.theta)
      done = False
    else:
      reward = 1.0
      done = np.abs(self.robot.theta) > .2
    self.rewards = [float(reward)]
    self.HUD(state, a, done)
    return state, sum(self.rewards), done, {}

  def camera_adjust(self):
    self.camera.move_and_look_at(0, 1.2, 1.0, 0, 0, 0.5)


class InvertedPendulumSwingupBulletEnv(InvertedPendulumBulletEnv):

  def __init__(self):
    self.robot = InvertedPendulumSwingup()
    MJCFBaseBulletEnv.__init__(self, self.robot)
    self.stateId = -1


class InvertedDoublePendulumBulletEnv(MJCFBaseBulletEnv):

  def __init__(self):
    self.robot = InvertedDoublePendulum()
    MJCFBaseBulletEnv.__init__(self, self.robot)
    self.stateId = -1

  def create_single_player_scene(self, bullet_client):
    return SingleRobotEmptyScene(bullet_client, gravity=9.8, timestep=0.0165, frame_skip=1)

  def reset(self):
    if (self.stateId >= 0):
      self._p.restoreState(self.stateId)
    r = MJCFBaseBulletEnv.reset(self)
    if (self.stateId < 0):
      self.stateId = self._p.saveState()
    return r

  def step(self, a):
    self.robot.apply_action(a)
    self.scene.global_step()
    state = self.robot.calc_state()  # sets self.pos_x self.pos_y
    # upright position: 0.6 (one pole) + 0.6 (second pole) * 0.5 (middle of second pole) = 0.9
    # using <site> tag in original xml, upright position is 0.6 + 0.6 = 1.2, difference +0.3
    dist_penalty = 0.01 * self.robot.pos_x**2 + (self.robot.pos_y + 0.3 - 2)**2
    # v1, v2 = self.model.data.qvel[1:3]   TODO when this fixed https://github.com/bulletphysics/bullet3/issues/1040
    #vel_penalty = 1e-3 * v1**2 + 5e-3 * v2**2
    vel_penalty = 0
    alive_bonus = 10
    done = self.robot.pos_y + 0.3 <= 1
    self.rewards = [float(alive_bonus), float(-dist_penalty), float(-vel_penalty)]
    self.HUD(state, a, done)
    return state, sum(self.rewards), done, {}

  def camera_adjust(self):
    self.camera.move_and_look_at(0, 1.2, 1.2, 0, 0, 0.5)



from pybullet_envs.bullet.cartpole_bullet import CartPoleBulletEnv
from pybullet_envs.bullet.cartpole_bullet import CartPoleContinuousBulletEnv
from pybullet_envs.bullet.minitaur_gym_env import MinitaurBulletEnv
from pybullet_envs.bullet.minitaur_duck_gym_env import MinitaurBulletDuckEnv
from pybullet_envs.bullet.racecarGymEnv import RacecarGymEnv
from pybullet_envs.bullet.racecarZEDGymEnv import RacecarZEDGymEnv
from pybullet_envs.bullet.kukaGymEnv import KukaGymEnv
from pybullet_envs.bullet.kukaCamGymEnv import KukaCamGymEnv
from pybullet_envs.bullet.kuka_diverse_object_gym_env import KukaDiverseObjectEnv



import pybullet as p
import time
import pybullet_data

p.connect(p.GUI)
p.setAdditionalSearchPath(pybullet_data.getDataPath())
useCollisionShapeQuery = True
p.configureDebugVisualizer(p.COV_ENABLE_GUI, 0)
geom = p.createCollisionShape(p.GEOM_SPHERE, radius=0.1)
geomBox = p.createCollisionShape(p.GEOM_BOX, halfExtents=[0.2, 0.2, 0.2])
baseOrientationB = p.getQuaternionFromEuler([0, 0.3, 0])  #[0,0.5,0.5,0]
basePositionB = [1.5, 0, 1]
obA = -1
obB = -1

obA = p.createMultiBody(baseMass=0, baseCollisionShapeIndex=geom, basePosition=[0.5, 0, 1])
obB = p.createMultiBody(baseMass=0,
                        baseCollisionShapeIndex=geomBox,
                        basePosition=basePositionB,
                        baseOrientation=baseOrientationB)

lineWidth = 3
colorRGB = [1, 0, 0]
lineId = p.addUserDebugLine(lineFromXYZ=[0, 0, 0],
                            lineToXYZ=[0, 0, 0],
                            lineColorRGB=colorRGB,
                            lineWidth=lineWidth,
                            lifeTime=0)
pitch = 0
yaw = 0

while (p.isConnected()):
  pitch += 0.01
  if (pitch >= 3.1415 * 2.):
    pitch = 0
  yaw += 0.01
  if (yaw >= 3.1415 * 2.):
    yaw = 0

  baseOrientationB = p.getQuaternionFromEuler([yaw, pitch, 0])
  if (obB >= 0):
    p.resetBasePositionAndOrientation(obB, basePositionB, baseOrientationB)

  if (useCollisionShapeQuery):
    pts = p.getClosestPoints(bodyA=-1,
                             bodyB=-1,
                             distance=100,
                             collisionShapeA=geom,
                             collisionShapeB=geomBox,
                             collisionShapePositionA=[0.5, 0, 1],
                             collisionShapePositionB=basePositionB,
                             collisionShapeOrientationB=baseOrientationB)
    #pts = p.getClosestPoints(bodyA=obA, bodyB=-1, distance=100, collisionShapeB=geomBox, collisionShapePositionB=basePositionB, collisionShapeOrientationB=baseOrientationB)
  else:
    pts = p.getClosestPoints(bodyA=obA, bodyB=obB, distance=100)

  if len(pts) > 0:
    #print(pts)
    distance = pts[0][8]
    #print("distance=",distance)
    ptA = pts[0][5]
    ptB = pts[0][6]
    p.addUserDebugLine(lineFromXYZ=ptA,
                       lineToXYZ=ptB,
                       lineColorRGB=colorRGB,
                       lineWidth=lineWidth,
                       lifeTime=0,
                       replaceItemUniqueId=lineId)
  #time.sleep(1./240.)

#removeCollisionShape is optional:
#only use removeCollisionShape if the collision shape is not used to create a body
#and if you want to keep on creating new collision shapes for different queries (not recommended)
p.removeCollisionShape(geom)
p.removeCollisionShape(geomBox)



import pybullet as p
import time
import pybullet_data

#Once the video is recorded, you can extract all individual frames using ffmpeg
#mkdir frames
#ffmpeg -i test.mp4 "frames/out-%03d.png"

#by default, PyBullet runs at 240Hz
p.connect(p.GUI, options="--width=320 --height=200 --mp4=\"test.mp4\" --mp4fps=240")

p.setAdditionalSearchPath(pybullet_data.getDataPath())
p.configureDebugVisualizer(p.COV_ENABLE_GUI,0)
p.configureDebugVisualizer(p.COV_ENABLE_SINGLE_STEP_RENDERING,1)
p.loadURDF("plane.urdf")

#in 3 seconds, the object travels about 0.5*g*t^2 meter ~ 45 meter.
r2d2 = p.loadURDF("r2d2.urdf",[0,0,45])
#disable linear damping
p.changeDynamics(r2d2,-1, linearDamping=0)
p.setGravity(0,0,-10)

for i in range (3*240):
  txt = "frame "+str(i)
  item = p.addUserDebugText(txt, [0,1,0])
  p.stepSimulation()
  #synchronize the visualizer (rendering frames for the video mp4) with stepSimulation
  p.configureDebugVisualizer(p.COV_ENABLE_SINGLE_STEP_RENDERING,1)
  #print("r2d2 vel=", p.getBaseVelocity(r2d2)[0][2])
  p.removeUserDebugItem(item)

p.disconnect()



"""Randomize the minitaur_gym_env when reset() is called."""
import random
import numpy as np
from . import env_randomizer_base

# Relative range.
MINITAUR_BASE_MASS_ERROR_RANGE = (-0.2, 0.2)  # 0.2 means 20%
MINITAUR_LEG_MASS_ERROR_RANGE = (-0.2, 0.2)  # 0.2 means 20%
# Absolute range.
BATTERY_VOLTAGE_RANGE = (14.8, 16.8)  # Unit: Volt
MOTOR_VISCOUS_DAMPING_RANGE = (0, 0.01)  # Unit: N*m*s/rad (torque/angular vel)
MINITAUR_LEG_FRICTION = (0.8, 1.5)  # Unit: dimensionless


class MinitaurEnvRandomizer(env_randomizer_base.EnvRandomizerBase):
  """A randomizer that change the minitaur_gym_env during every reset."""

  def __init__(self,
               minitaur_base_mass_err_range=MINITAUR_BASE_MASS_ERROR_RANGE,
               minitaur_leg_mass_err_range=MINITAUR_LEG_MASS_ERROR_RANGE,
               battery_voltage_range=BATTERY_VOLTAGE_RANGE,
               motor_viscous_damping_range=MOTOR_VISCOUS_DAMPING_RANGE):
    self._minitaur_base_mass_err_range = minitaur_base_mass_err_range
    self._minitaur_leg_mass_err_range = minitaur_leg_mass_err_range
    self._battery_voltage_range = battery_voltage_range
    self._motor_viscous_damping_range = motor_viscous_damping_range

  def randomize_env(self, env):
    self._randomize_minitaur(env.minitaur)

  def _randomize_minitaur(self, minitaur):
    """Randomize various physical properties of minitaur.

    It randomizes the mass/inertia of the base, mass/inertia of the legs,
    friction coefficient of the feet, the battery voltage and the motor damping
    at each reset() of the environment.

    Args:
      minitaur: the Minitaur instance in minitaur_gym_env environment.
    """
    base_mass = minitaur.GetBaseMassFromURDF()
    randomized_base_mass = random.uniform(
        base_mass * (1.0 + self._minitaur_base_mass_err_range[0]),
        base_mass * (1.0 + self._minitaur_base_mass_err_range[1]))
    minitaur.SetBaseMass(randomized_base_mass)

    leg_masses = minitaur.GetLegMassesFromURDF()
    leg_masses_lower_bound = np.array(leg_masses) * (1.0 + self._minitaur_leg_mass_err_range[0])
    leg_masses_upper_bound = np.array(leg_masses) * (1.0 + self._minitaur_leg_mass_err_range[1])
    randomized_leg_masses = [
        np.random.uniform(leg_masses_lower_bound[i], leg_masses_upper_bound[i])
        for i in range(len(leg_masses))
    ]
    minitaur.SetLegMasses(randomized_leg_masses)

    randomized_battery_voltage = random.uniform(BATTERY_VOLTAGE_RANGE[0], BATTERY_VOLTAGE_RANGE[1])
    minitaur.SetBatteryVoltage(randomized_battery_voltage)

    randomized_motor_damping = random.uniform(MOTOR_VISCOUS_DAMPING_RANGE[0],
                                              MOTOR_VISCOUS_DAMPING_RANGE[1])
    minitaur.SetMotorViscousDamping(randomized_motor_damping)

    randomized_foot_friction = random.uniform(MINITAUR_LEG_FRICTION[0], MINITAUR_LEG_FRICTION[1])
    minitaur.SetFootFriction(randomized_foot_friction)



# Lint as: python3
"""A module that defines autonomous object class and related functions."""
from typing import Any, Callable, Dict, Optional, Sequence, Text, Union

from absl import logging
import gin
import numpy as np

from pybullet_utils import bullet_client
from pybullet_envs.minitaur.envs_v2 import base_client
from pybullet_envs.minitaur.envs_v2.sensors import sensor
from pybullet_envs.minitaur.robots import object_controller
from pybullet_envs.minitaur.robots import robot_base

# The action value to pass into AutonomousObject pre_control_step() and
# apply_action().
AUTONOMOUS_ACTION = None

# Maximum force used in constraint based actuation.
_MAX_FORCE = 1000


# TODO(b/155124699): find a better way actuate object than using constraint or
# modifying URDF.
@gin.configurable
class AutonomousObject(robot_base.RobotBase):
  """Autonomous object that moves/acts in simulation guided by a controller."""

  def __init__(self,
               urdf_file: Text,
               sensors: Sequence[sensor.Sensor] = (),
               controller: object_controller.ControllerBase = None,
               actuate_by_reset: bool = False):
    """Constructor.

    Args:
      urdf_file: The path to urdf file of the object.
      sensors: A list of sensor objects to attach to autonomous object.
      controller: A controller object that governs autonomous object's motion.
        If not specified, StationaryController is used.
      actuate_by_reset: Use pybullet resetBasePositionAndOrientation to actuate
        the object. Default is False, which means actuate by constraint. In the
        actuate by constrained mode, be extra cautious when the position or
        orientation control is based on position or orientation sensor reading
        of the same object. This loop-back condition is known to be problematic
        and causes slower than expected motion.
    """
    self._urdf_file = urdf_file
    self._controller = controller or object_controller.StationaryController()
    self._actuate_by_reset = actuate_by_reset
    self._actuate_function = (
        self._actuate_base_pose
        if not actuate_by_reset else self._reset_base_pose)

    self._sensors = list(sensors)
    self._object_id = -1
    self._constraint_id = -1

    self._pybullet_client = None  # will be initialized in set_sim_client()
    self._clock = None  # will be initialized in set_clock()
    self._init_internal_states()

  def _init_internal_states(self) -> None:
    self._observations_time_since_reset = 0
    self._observations = {}
    self._position = np.zeros(3)
    self._orientation = np.array([0, 0, 0, 1])

  def set_sim_client(self, pybullet_client: bullet_client.BulletClient) -> None:
    """Sets new simulation client and reload assets."""
    self._pybullet_client = pybullet_client
    self._init_internal_states()
    self.load()

  def set_clock(self, clock: Callable[[], float]) -> None:
    """Sets monotonic clock when adding into simulation environment."""
    self._clock = clock

  @property
  def sim_object_id(self):
    return self._object_id

  def update(self, time_since_reset_sec: float,
             observations: Dict[Text, Any]) -> None:
    """Updates simulation time and observations.

    This function should be called before apply_action in each simulation step.

    Args:
      time_since_reset_sec: Time from start of simulation reset in seconds.
      observations: A dict of observations.
    """
    if time_since_reset_sec < self._observations_time_since_reset:
      raise ValueError(
          "Time cannot go backwards. Current t = %f, new t = %f." %
          (self._observations_time_since_reset, time_since_reset_sec))
    self._observations_time_since_reset = time_since_reset_sec
    self._observations = observations

  def _load_urdf(self):
    """Loads object URDF file."""
    try:
      print("loading: ", self._urdf_file)
      self._object_id = self._pybullet_client.loadURDF(self._urdf_file)
    except:
      print("Error: cannot load ", self._urdf_file)
      import sys
      sys.exit(0)

  def load(self) -> None:
    """Reconstructs the robot and resets its states."""
    self._load_urdf()
    if not self._actuate_by_reset:
      self._constraint_id = self._pybullet_client.createConstraint(
          parentBodyUniqueId=self._object_id,
          parentLinkIndex=-1,
          childBodyUniqueId=-1,
          childLinkIndex=-1,
          jointType=self._pybullet_client.JOINT_FIXED,
          jointAxis=(0, 0, 0),
          parentFramePosition=(0, 0, 0),
          childFramePosition=(0, 0, 0),
          childFrameOrientation=(0, 0, 0, 1))

    for s in self._sensors:
      s.set_robot(self)

    # Resets the pose and updates the initial observations.
    self.reset()

  def reset(
      self,
      base_position: Optional[Sequence[float]] = None,
      base_orientation_quaternion: Optional[Sequence[float]] = None,
      controller: Optional[object_controller.ControllerBase] = None) -> None:
    """Resets the states (e.g.

    pose and sensor readings) of the robot.

    This is called at the start of each episode by the environment.

    Args:
      base_position: Robot base position after reset. Must be None.
      base_orientation_quaternion: Robot base orientation after reset. Must be
        None.
      controller: A new controller to replace original controller.
    """
    if base_position is not None or base_orientation_quaternion is not None:
      raise ValueError("Reset position and orientation of AutonomousObject is "
                       "specified in controller.")

    if controller is not None:
      self._controller = controller

    self._init_internal_states()
    self._position, self._orientation, _ = self._controller.get_action(
        object_controller.INIT_TIME, self._observations)

    self._reset_base_pose(self._position, self._orientation)
    self.receive_observation()

  def terminate(self) -> None:
    """Shuts down the robot, no-op in simulation."""

  def pre_control_step(self, action: Any) -> Any:
    """Processes the input action before the action repeat loop.

    Args:
      action: expect it to be `AUTONOMOUS_ACTION` at present.

    Returns:
      the action as is.
    """
    # Environment should not pass action other than AUTONOMOUS_ACTION.
    if action is not AUTONOMOUS_ACTION:
      raise ValueError("AutonomousObject only accept AUTONOMOUS_ACTION as "
                       "action value input.")
    return action

  def apply_action(self, action: Any) -> None:
    """Applies the action to the robot."""
    # Environment should not pass action other than AUTONOMOUS_ACTION.
    if action is not AUTONOMOUS_ACTION:
      raise ValueError("AutonomousObject only accept AUTONOMOUS_ACTION as "
                       "action value input.")
    position, orientation, _ = self._controller.get_action(
        self._observations_time_since_reset, self._observations)

    self._actuate_function(position, orientation)

  def receive_observation(self) -> None:
    """Updates the robot sensor readings."""
    position, orientation = (
        self._pybullet_client.getBasePositionAndOrientation(self._object_id))
    self._position = np.array(position)
    self._orientation = np.array(orientation)

  def post_control_step(self) -> None:
    """Updates internal variables. Not yet used in AutonomousObject."""
    pass

  def _reset_base_pose(self,
                       position: Union[Sequence[float], np.ndarray] = None,
                       orientation_quat: Union[Sequence[float],
                                               np.ndarray] = None):
    """Resets the base to the desired position and orientation.

    Args:
      position: The desired base position. If omitted, current location is used.
      orientation_quat: The desired base orientation in quaternion. If omitted,
        current orientation is used.
    """
    if position is None:
      position = self._position

    if orientation_quat is None:
      orientation_quat = self._orientation

    self._pybullet_client.resetBaseVelocity(self._object_id, (0, 0, 0),
                                            (0, 0, 0))
    self._pybullet_client.resetBasePositionAndOrientation(
        self._object_id, position, orientation_quat)

  def _actuate_base_pose(self, position: Union[Sequence[float], np.ndarray],
                         orientation_quat: Union[Sequence[float], np.ndarray]):
    """Actuates the base to the desired position and orientation.

    Difference of this function from _reset_base_pose() is that this function
    considers dynamics along the path and collisions along the motion path.

    Args:
      position: The desired base position.
      orientation_quat: The desired base orientation in quaternion.
    """
    self._pybullet_client.changeConstraint(
        self._constraint_id,
        position,
        jointChildFrameOrientation=orientation_quat,
        maxForce=_MAX_FORCE)

  def _reset_joint_angles(self, joint_angles=None):
    """Resets the joint angles. Not yet used in AutonomousObject."""
    del joint_angles

  @property
  def action_names(self) -> Sequence[Text]:
    """Returns a sequence of action names. Always () for AutonomousObject."""
    return ()

  @property
  def sensors(self) -> Sequence[sensor.Sensor]:
    """Returns the sensors on this robot."""
    return self._sensors

  @property
  def base_orientation_quaternion(self) -> np.ndarray:
    """Returns the base pose as a quaternion in format (x, y, z, w)."""
    return self._orientation.copy()

  @property
  def base_roll_pitch_yaw(self) -> np.ndarray:
    """Returns the base roll, pitch, and yaw angles in radians."""
    return np.array(
        self._pybullet_client.getEulerFromQuaternion(self._orientation))

  @property
  def base_position(self) -> np.ndarray:
    """Returns the base cartesian coordinates in meters."""
    return self._position.copy()

  @property
  def timestamp(self):
    """Simulation monotonic time."""
    if self._clock is None:
      raise RuntimeError("Must call set_clock() before accessing timestamp.")
    return self._clock()

  # This is need for CameraSensor.set_robot() to work.
  @property
  def pybullet_client(self):
    return self._pybullet_client

  # This is need for CameraSensor.set_robot() to work.
  @property
  def robot_id(self) -> int:
    return self._object_id



import pybullet as p
import pybullet_data as pd
import time
p.connect(p.GUI)
p.setAdditionalSearchPath(pd.getDataPath())
offset = 0
for scale in range (1,10,1):
  ball = p.loadURDF("soccerball.urdf",[offset,0,1], globalScaling=scale*0.1)
  p.changeDynamics(ball,-1,linearDamping=0, angularDamping=0, rollingFriction=0.001, spinningFriction=0.001)
  p.changeVisualShape(ball,-1,rgbaColor=[0.8,0.8,0.8,1])
  offset += 2*scale*0.1
p.loadURDF("plane.urdf")
p.setGravity(0,0,-10)
p.setRealTimeSimulation(1)
while p.isConnected():
  #p.getCameraImage(320,200, renderer=p.ER_BULLET_HARDWARE_OPENGL)
  time.sleep(0.5)



# Lint as: python3
"""This file implements the locomotion gym env."""

import atexit
import collections
import time
from typing import Any, Callable, Sequence, Text, Union
import gin
import gym
import numpy as np

from pybullet_envs.minitaur.envs_v2 import base_client
from pybullet_utils import bullet_client
import pybullet_data
import pybullet
from pybullet_envs.minitaur.envs import minitaur_logging
from pybullet_envs.minitaur.envs import minitaur_logging_pb2
#from pybullet_envs.minitaur.envs import minitaur_logging
#from pybullet_envs.minitaur.envs import minitaur_logging_pb2
from pybullet_envs.minitaur.envs_v2.evaluation import metric_logger
from pybullet_envs.minitaur.envs_v2.scenes import scene_base
from pybullet_envs.minitaur.envs_v2.sensors import sensor
from pybullet_envs.minitaur.envs_v2.sensors import space_utils
from pybullet_envs.minitaur.envs_v2.utilities import rendering_utils
from pybullet_envs.minitaur.robots import autonomous_object
from pybullet_envs.minitaur.robots import robot_base

_ACTION_EPS = 0.01
_NUM_SIMULATION_ITERATION_STEPS = 300
_LOG_BUFFER_LENGTH = 5000

SIM_CLOCK = 'SIM_CLOCK'

# Exports this symbol so we can use it in the config file.
gin.constant('locomotion_gym_env.SIM_CLOCK', SIM_CLOCK)

# This allows us to bind @time.time in the gin configuration.
gin.external_configurable(time.time, module='time')


# TODO(b/122048194): Enable position/torque/hybrid control mode.
@gin.configurable
class LocomotionGymEnv(gym.Env):
  """The gym environment for the locomotion tasks."""
  metadata = {
      'render.modes': ['human', 'rgb_array', 'topdown'],
      'video.frames_per_second': 100
  }

  def __init__(self,
               gym_config,
               clock: Union[Callable[..., float], Text] = 'SIM_CLOCK',
               robot_class: Any = None,
               scene: scene_base.SceneBase = None,
               sensors: Sequence[sensor.Sensor] = None,
               task: Any = None,
               env_randomizers: Any = None):
    """Initializes the locomotion gym environment.

    Args:
      gym_config: An instance of LocomotionGymConfig.
      clock: The clock source to be used for the gym env. The clock should
        return a timestamp in seconds. Setting clock == "SIM_CLOCK" will enable
        the built-in simulation clock. For real robot experiments, we can use
        time.time or other clock wall clock sources.
      robot_class: A class of a robot. We provide a class rather than an
        instance due to hard_reset functionality. Parameters are expected to be
        configured with gin.
      scene: An object for managing the robot's surroundings.
      sensors: A list of environmental sensors for observation.
      task: A callable function/class to calculate the reward and termination
        condition. Takes the gym env as the argument when calling.
      env_randomizers: A list of EnvRandomizer(s). An EnvRandomizer may
        randomize the physical property of minitaur, change the terrrain during
        reset(), or add perturbation forces during step().
      client_factory: A function to create a simulation client, it can be a
        pybullet client.

    Raises:
      ValueError: If the num_action_repeat is less than 1.

    """
    self._pybullet_client = None
    self._metric_logger = metric_logger.MetricLogger()
    # TODO(sehoonha) split observation and full-state sensors (b/129858214)

    # Makes sure that close() is always called to flush out the logs to the
    # disk.
    atexit.register(self.close)
    self.seed()
    self._gym_config = gym_config
    if robot_class is None:
      raise ValueError('robot_class cannot be None.')
    self._robot_class = robot_class
    if issubclass(self._robot_class, robot_base.RobotBase):
      self._use_new_robot_class = True
    else:
      self._use_new_robot_class = False
    self._robot = None

    self._scene = scene or scene_base.SceneBase()

    # TODO(sehoonha) change the data structure to dictionary
    self._env_sensors = list(sensors) if sensors is not None else list()

    # TODO(b/152161457): Make logging a standalone module.
    self._log_path = gym_config.log_path
    self._logging = minitaur_logging.MinitaurLogging(self._log_path)
    self._episode_proto = minitaur_logging_pb2.MinitaurEpisode()
    self._data_dir = gym_config.data_dir

    self._task = task

    self._env_randomizers = env_randomizers if env_randomizers else []

    # Simulation related parameters.
    self._num_action_repeat = gym_config.simulation_parameters.num_action_repeat
    self._on_rack = gym_config.simulation_parameters.robot_on_rack
    if self._num_action_repeat < 1:
      raise ValueError('number of action repeats should be at least 1.')
    # TODO(b/73829334): Fix the value of self._num_bullet_solver_iterations.
    self._num_bullet_solver_iterations = int(_NUM_SIMULATION_ITERATION_STEPS /
                                             self._num_action_repeat)

    self._sim_time_step = gym_config.simulation_parameters.sim_time_step_s
    # The sim step counter is an internal varialbe to count the number of
    # pybullet stepSimulation() has been called since last reset.
    self._sim_step_counter = 0

    self._env_time_step = self._num_action_repeat * self._sim_time_step
    # The env step counter accounts for how many times env.step has been
    # called since reset.
    self._env_step_counter = 0

    if clock == SIM_CLOCK:
      self._clock = self._get_sim_time
    else:
      self._clock = clock

    # Creates the bullet client.
    self._is_render = gym_config.simulation_parameters.enable_rendering
    # The wall-clock time at which the last frame is rendered.
    self._last_frame_time = 0.0
    
    if gym_config.simulation_parameters.enable_rendering:
      self._pybullet_client = bullet_client.BulletClient(connection_mode=pybullet.GUI)
      self._pybullet_client.configureDebugVisualizer(
        pybullet.COV_ENABLE_GUI,
        gym_config.simulation_parameters.enable_rendering_gui)
    else:
      self._pybullet_client = bullet_client.BulletClient()
    if gym_config.simulation_parameters.egl_rendering:
      self._pybullet_client.loadPlugin('eglRendererPlugin')

    self._pybullet_client.setAdditionalSearchPath(
      pybullet_data.getDataPath())

    # If enabled, save the performance profile to profiling_path
    # Use Google Chrome about://tracing to open the file
    if gym_config.profiling_path is not None:
      self._profiling_slot = self._pybullet_client.startStateLogging(
          self._pybullet_client.STATE_LOGGING_PROFILE_TIMINGS,
          gym_config.profiling_path)
      self._profiling_counter = 10
    else:
      self._profiling_slot = -1

    # Set the default render options. TODO(b/152161124): Make rendering a
    # standalone module.
    self._camera_target = gym_config.simulation_parameters.camera_target
    self._camera_dist = gym_config.simulation_parameters.camera_distance
    self._camera_yaw = gym_config.simulation_parameters.camera_yaw
    self._camera_pitch = gym_config.simulation_parameters.camera_pitch
    self._render_width = gym_config.simulation_parameters.render_width
    self._render_height = gym_config.simulation_parameters.render_height

    # Loads the environment and robot. Actions space will be created as well.
    self._hard_reset = True
    self._observation_dict = {}
    self.reset()
    self._hard_reset = gym_config.simulation_parameters.enable_hard_reset

    # Construct the observation space from the list of sensors.
    self.observation_space = (
        space_utils.convert_sensors_to_gym_space_dictionary([
            sensor for sensor in self.all_sensors()
            if sensor.get_name() not in self._gym_config.ignored_sensor_list
        ]))

  def __del__(self):
    self.close()

  def _load_old_robot_class(self):
    self._robot = self._robot_class(
        pybullet_client=self._pybullet_client, on_rack=self._on_rack)
    self._action_list = []
    action_upper_bound = []
    action_lower_bound = []
    for action in self._gym_config.actions:
      self._action_list.append(action.name)
      action_upper_bound.append(action.upper_bound)
      action_lower_bound.append(action.lower_bound)
    self.action_space = gym.spaces.Box(
        np.array(action_lower_bound),
        np.array(action_upper_bound),
        dtype=np.float32)

  def _load_new_robot_class(self):
    self._robot = self._robot_class(
        pybullet_client=self._pybullet_client, clock=self._clock)
    self.action_space = self._robot.action_space

  def _load(self):
    self._pybullet_client.resetSimulation()
    self._pybullet_client.setPhysicsEngineParameter(
        numSolverIterations=self._num_bullet_solver_iterations)
    self._pybullet_client.setTimeStep(self._sim_time_step)
    self._pybullet_client.setGravity(0, 0, -10)
    self._pybullet_client.setPhysicsEngineParameter(enableConeFriction=0)

    # Disable rendering during scene loading will speed up simulation.
    if self._is_render:
      self._pybullet_client.configureDebugVisualizer(
          self._pybullet_client.COV_ENABLE_RENDERING, 0)

    # Rebuild the scene.
    self._scene.build_scene(self._pybullet_client)

    # TODO(b/151975607): Deprecate old robot support.
    if self._use_new_robot_class:
      self._load_new_robot_class()
    else:
      self._load_old_robot_class()

    # Check action space.
    if (isinstance(self.action_space, gym.spaces.Box) and
        not np.all(self.action_space.low < self.action_space.high)):
      raise ValueError(f'Action space contains invalid dimensions, '
                       f'action space low = {self.action_space.low}, '
                       f'action space high = {self.action_space.high}')

    for an_object in self._dynamic_objects():
      an_object.set_clock(self._clock)

    # Enable rendering after loading finishes.
    if self._is_render:
      self._pybullet_client.configureDebugVisualizer(
          self._pybullet_client.COV_ENABLE_RENDERING, 1)

  def close(self):
    atexit.unregister(self.close)
    #if self._pybullet_client:
    
    if self._log_path is not None:
      self._logging.save_episode(self._episode_proto)
    for sensor_ in self.all_sensors():
      sensor_.on_terminate(self)
    if self._robot:
      if self._use_new_robot_class:
        self._robot.terminate()
      else:
        self._robot.Terminate()
    if self._pybullet_client:
      self._pybullet_client.disconnect()
      self._pybullet_client = None

  def seed(self, seed=None):
    self.np_random, self.np_random_seed = gym.utils.seeding.np_random(seed)
    return [self.np_random_seed]

  def _dynamic_objects(self):
    """Returns the python objects controlling moving obstacles."""
    if self._scene:
      return self._scene.dynamic_objects
    else:
      return []

  def all_sensors(self):
    """Returns all robot, environmental and dynamic objects sensors."""
    if self._use_new_robot_class:
      all_sensors = list(self._env_sensors)
      if self._robot:
        all_sensors.extend(list(self._robot.sensors))
      for obj in self._dynamic_objects():
        all_sensors.extend(obj.sensors)

      # The new way of adding task specific sensors to the sensor lists.
      if hasattr(self._task, 'sensors'):
        all_sensors.extend(self._task.sensors)
      return all_sensors
    else:
      # This is a workaround due to the issue in b/130128505#comment5
      task_sensor = ([self._task]
                     if isinstance(self._task, sensor.Sensor) else [])
      robot_sensors = []
      if self._robot:
        robot_sensors = self._robot.GetAllSensors()
      return robot_sensors + self._env_sensors + task_sensor

  def sensor_by_name(self, name):
    """Returns the sensor with the given name, or None if not exist."""
    # TODO(b/154162104): Store sensors as dictionary.
    for sensor_ in self.all_sensors():
      if sensor_.get_name() == name:
        return sensor_
    return None

  @gin.configurable('locomotion_gym_env.LocomotionGymEnv.reset')
  def reset(
      self,
      initial_motor_angles=None,
      reset_duration=1.0,
      reset_visualization_camera=True,
  ):
    """Resets the robot's position in the world or rebuild the sim world.

    The simulation world will be rebuilt if self._hard_reset is True.

    Args:
      initial_motor_angles: A list of Floats. The desired joint angles after
        reset. If None, the robot will use its built-in value.
      reset_duration: Float. The time (in seconds) needed to rotate all motors
        to the desired initial values.
      reset_visualization_camera: Whether to reset debug visualization camera on
        reset.

    Returns:
      A numpy array contains the initial observation after reset.
    """
    

    self._env_step_counter = 0
    self._sim_step_counter = 0
    self._last_reset_time = self._clock()
    self._metric_logger.reset_episode()

    # Clear the simulation world and rebuild the robot interface.
    if self._hard_reset:
      self._load()

    # Resets the scene
    self._scene.reset()

    # Resets the robot with the provided init parameters.
    if self._use_new_robot_class:
      self._robot.reset()
    else:
      self._robot.Reset(
          reload_urdf=False,
          default_motor_angles=initial_motor_angles,
          reset_time=reset_duration)

    # Flush the logs to disc and reinitialize the logging system.
    if self._log_path is not None:
      self._logging.save_episode(self._episode_proto)
      self._episode_proto = minitaur_logging_pb2.MinitaurEpisode()
      minitaur_logging.preallocate_episode_proto(self._episode_proto,
                                                 _LOG_BUFFER_LENGTH,
                                                 self._robot)

    # TODO(b/152161124): Move this part to the renderer module.
    if reset_visualization_camera:
      self._pybullet_client.resetDebugVisualizerCamera(self._camera_dist,
                                                       self._camera_yaw,
                                                       self._camera_pitch,
                                                       [0, 0, 0])

    # Create an example last action based on the type of action space.
    self._last_action = space_utils.create_constant_action(self.action_space)

    for s in self.all_sensors():
      s.on_reset(self)

    if self._task and hasattr(self._task, 'reset'):
      self._task.reset(self)

    # Loop over all env randomizers.
    for env_randomizer in self._env_randomizers:
      env_randomizer.randomize_env(self)

    for obj in self._dynamic_objects():
      obj.reset()

    # Initialize the robot base position.
    if self._use_new_robot_class:
      self._last_base_position = self._robot.base_position
    else:
      self._last_base_position = self._robot.GetBasePosition()

    # Resets the observations again, since randomizers might change the env.
    for s in self.all_sensors():
      s.on_reset(self)

    
    self._last_reset_time = self._clock()
    return self._get_observation()

  def _wait_for_rendering(self):
    # Sleep, otherwise the computation takes less time than real time,
    # which will make the visualization like a fast-forward video.
    time_spent = time.time() - self._last_frame_time
    self._last_frame_time = time.time()
    time_to_sleep = self._env_time_step - time_spent
    if time_to_sleep > 0:
      time.sleep(time_to_sleep)

    # Also keep the previous orientation of the camera set by the user.
    [yaw, pitch, dist] = self._pybullet_client.getDebugVisualizerCamera()[8:11]
    self._pybullet_client.resetDebugVisualizerCamera(dist, yaw, pitch,
                                                     self._last_base_position)

  def _step_old_robot_class(self, action):
    self._last_base_position = self._robot.GetBasePosition()
    self._last_action = action

    if self._is_render:
      self._wait_for_rendering()

    for env_randomizer in self._env_randomizers:
      env_randomizer.randomize_step(self)

    self._robot.Step(action)

    if self._profiling_slot >= 0:
      self._profiling_counter -= 1
      if self._profiling_counter == 0:
        self._pybullet_client.stopStateLogging(self._profiling_slot)

    if self._log_path is not None:
      minitaur_logging.update_episode_proto(self._episode_proto, self._robot,
                                            action, self._env_step_counter)
    reward = self._reward()

    for s in self.all_sensors():
      s.on_step(self)

    if self._task and hasattr(self._task, 'update'):
      self._task.update(self)  # TODO(b/154635313): resolve API mismatch

    done = self._termination()
    self._env_step_counter += 1
    # TODO(b/161941829): terminate removed for now, change terminate to other
    # names.
    return self._get_observation(), reward, done, {}

  def _step_new_robot_class(self, action):
    self._last_base_position = self._robot.base_position
    self._last_action = action

    if self._is_render:
      self._wait_for_rendering()

    for env_randomizer in self._env_randomizers:
      env_randomizer.randomize_step(self)

    action = self._robot.pre_control_step(action, self._env_time_step)
    for obj in self._dynamic_objects():
      obj.pre_control_step(autonomous_object.AUTONOMOUS_ACTION)
    for _ in range(self._num_action_repeat):
      self._robot.apply_action(action)
      for obj in self._dynamic_objects():
        obj.update(self.get_time_since_reset(), self._observation_dict)
        obj.apply_action(autonomous_object.AUTONOMOUS_ACTION)

      self._pybullet_client.stepSimulation()
      self._sim_step_counter += 1

      self._robot.receive_observation()
      for obj in self._dynamic_objects():
        obj.receive_observation()

      for s in self.all_sensors():
        s.on_new_observation()

    self._robot.post_control_step()
    for obj in self._dynamic_objects():
      obj.post_control_step()

    if self._profiling_slot >= 0:
      self._profiling_counter -= 1
      if self._profiling_counter == 0:
        self._pybullet_client.stopStateLogging(self._profiling_slot)

    if self._log_path is not None:
      minitaur_logging.update_episode_proto(self._episode_proto, self._robot,
                                            action, self._env_step_counter)
    reward = self._reward()

    for s in self.all_sensors():
      s.on_step(self)

    if self._task and hasattr(self._task, 'update'):
      self._task.update(self)  # TODO(b/154635313): resolve API mismatch

    done = self._termination()
    self._env_step_counter += 1
    # TODO(b/161941829): terminate removed for now, change terminate to other
    # names.
    return self._get_observation(), reward, done, {}

  def step(self, action):
    """Step forward the simulation, given the action.

    Args:
      action: Can be a list of desired motor angles for all motors when the
        robot is in position control mode; A list of desired motor torques. Or a
        list of tuples (q, qdot, kp, kd, tau) for hybrid control mode. The
        action must be compatible with the robot's motor control mode. Also, we
        are not going to use the leg space (swing/extension) definition at the
        gym level, since they are specific to Minitaur.

    Returns:
      observations: The observation dictionary. The keys are the sensor names
        and the values are the sensor readings.
      reward: The reward for the current state-action pair.
      done: Whether the episode has ended.
      info: A dictionary that stores diagnostic information.

    Raises:
      ValueError: The action dimension is not the same as the number of motors.
      ValueError: The magnitude of actions is out of bounds.
    """
    # TODO(b/151975607): Finish the migration and remove old robot class
    # support.
    if self._use_new_robot_class:
      return self._step_new_robot_class(action)
    else:
      return self._step_old_robot_class(action)

  @gin.configurable('locomotion_gym_env.LocomotionGymEnv.render')
  def render(self, mode='rgb_array'):
    
    if mode == 'topdown':
      # Provide ground height if we know it. Otherwise leave it as gin
      # configurable.
      if hasattr(self.scene, 'ground_height'):
        return rendering_utils.render_topdown(
            self._pybullet_client, ground_height=self.scene.ground_height)
      else:
        return rendering_utils.render_topdown(self._pybullet_client)

    if mode != 'rgb_array':
      raise ValueError('Unsupported render mode:{}'.format(mode))

    if self._camera_target is not None:
      target_position = self._camera_target
    else:
      target_position = self._last_base_position
    view_matrix = self._pybullet_client.computeViewMatrixFromYawPitchRoll(
        cameraTargetPosition=target_position,
        distance=self._camera_dist,
        yaw=self._camera_yaw,
        pitch=self._camera_pitch,
        roll=0,
        upAxisIndex=2)
    proj_matrix = self._pybullet_client.computeProjectionMatrixFOV(
        fov=60,
        aspect=float(self._render_width) / self._render_height,
        nearVal=0.1,
        farVal=100.0)
    return rendering_utils.render_image(self._pybullet_client,
                                        self._render_width, self._render_height,
                                        view_matrix, proj_matrix)

  @property
  def scene(self):
    return self._scene

  @property
  def rendering_enabled(self):
    return self._is_render

  @property
  def env_randomizers(self):
    return self._env_randomizers

  @property
  def last_base_position(self):
    return self._last_base_position

  @property
  def gym_config(self):
    return self._gym_config

  def _termination(self):
    if not self._robot.is_safe:
      return True

    if self._task and hasattr(self._task, 'done'):
      return self._task.done(self)  # TODO(b/154635313): resolve API mismatch

    return False

  def _reward(self):
    if self._task:
      return self._task.reward(self)  # TODO(b/154635313): resolve API mismatch
    return 0

  def _get_observation(self):
    """Get observation of this environment from a list of sensors.

    Returns:
      observations: dictionary of sensory observation with sensor name as key
        and corresponding observation in numpy array as value.
    """
    sensors_dict = {}
    for s in self.all_sensors():
      if s.get_name() in self._gym_config.ignored_sensor_list:
        continue

      obs = s.get_observation()
      if isinstance(obs, dict):
        sensors_dict.update(obs)
      else:
        sensors_dict[s.get_name()] = obs

    self._observation_dict = collections.OrderedDict(
        sorted(sensors_dict.items()))
    return self._observation_dict

  def set_time_step(self, num_action_repeat, sim_step=0.001):
    """Sets the time step of the environment.

    Args:
      num_action_repeat: The number of simulation steps/action repeats to be
        executed when calling env.step().
      sim_step: The simulation time step in PyBullet. By default, the simulation
        step is 0.001s, which is a good trade-off between simulation speed and
        accuracy.

    Raises:
      ValueError: If the num_action_repeat is less than 1.
    """
    if num_action_repeat < 1:
      raise ValueError('number of action repeats should be at least 1.')
    self._sim_time_step = sim_step
    self._num_action_repeat = num_action_repeat
    self._env_time_step = sim_step * num_action_repeat
    self._num_bullet_solver_iterations = int(
        _NUM_SIMULATION_ITERATION_STEPS / self._num_action_repeat)
    self._pybullet_client.setPhysicsEngineParameter(
        numSolverIterations=self._num_bullet_solver_iterations)
    self._pybullet_client.setTimeStep(self._sim_time_step)
    if not self._use_new_robot_class:
      self._robot.SetTimeSteps(self._num_action_repeat, self._sim_time_step)

  def _get_sim_time(self):
    """Returns the simulation time since the sim resets."""
    return self._sim_step_counter * self._sim_time_step

  def get_time_since_reset(self):
    """Get the time passed (in seconds) since the last reset.

    Returns:
      Time in seconds since the last reset.
    """
    if self._use_new_robot_class:
      return self._clock() - self._last_reset_time
    else:
      return self._robot.GetTimeSinceReset()

  def get_time(self):
    """Gets the time reading from the clock source."""
    return self._clock()

  @property
  def observation(self):
    return self._observation_dict

  @property
  def pybullet_client(self):
    return self._pybullet_client

  @property
  def robot(self):
    return self._robot

  @property
  def num_action_repeat(self):
    return self._num_action_repeat

  @property
  def sim_time_step(self):
    return self._sim_time_step

  @property
  def env_step_counter(self):
    return self._env_step_counter

  @property
  def hard_reset(self):
    return self._hard_reset

  @property
  def last_action(self):
    return self._last_action

  @property
  def env_time_step(self):
    return self._env_time_step

  @property
  def data_dir(self):
    return self._data_dir

  @property
  def task(self):
    return self._task

  @property
  def robot_class(self):
    return self._robot_class

  @property
  def action_names(self):
    """Name of each action in the action space.

    By default this returns the actions the robot executes (e.g.
    "VELOCITY_elbow_joint"), but env wrappers may override this if they change
    the action space (e.g. if they convert twist to motor commands).

    Returns:
      Tuple of strings, the action names.
    """
    if self._use_new_robot_class:
      return self._robot.action_names
    return self._action_list

  @property
  def metric_logger(self):
    return self._metric_logger



import numpy as np


class PDControllerStableMultiDof(object):

  def __init__(self, pb):
    self._pb = pb

  def computeAngVel(self, ornStart, ornEnd, deltaTime, bullet_client):
    dorn = bullet_client.getDifferenceQuaternion(ornStart, ornEnd)
    axis, angle = bullet_client.getAxisAngleFromQuaternion(dorn)
    angVel = [(axis[0] * angle) / deltaTime, (axis[1] * angle) / deltaTime,
              (axis[2] * angle) / deltaTime]
    return angVel

  def quatMul(self, q1, q2):
    return [
        q1[3] * q2[0] + q1[0] * q2[3] + q1[1] * q2[2] - q1[2] * q2[1],
        q1[3] * q2[1] + q1[1] * q2[3] + q1[2] * q2[0] - q1[0] * q2[2],
        q1[3] * q2[2] + q1[2] * q2[3] + q1[0] * q2[1] - q1[1] * q2[0],
        q1[3] * q2[3] - q1[0] * q2[0] - q1[1] * q2[1] - q1[2] * q2[2]
    ]

  def computeAngVelRel(self, ornStart, ornEnd, deltaTime, bullet_client):
    ornStartConjugate = [-ornStart[0], -ornStart[1], -ornStart[2], ornStart[3]]
    q_diff = self.quatMul(
        ornStartConjugate,
        ornEnd)  #bullet_client.multiplyTransforms([0,0,0], ornStartConjugate, [0,0,0], ornEnd)

    axis, angle = bullet_client.getAxisAngleFromQuaternion(q_diff)
    angVel = [(axis[0] * angle) / deltaTime, (axis[1] * angle) / deltaTime,
              (axis[2] * angle) / deltaTime]
    return angVel

  def computePD(self, bodyUniqueId, jointIndices, desiredPositions, desiredVelocities, kps, kds,
                maxForces, timeStep):

    numJoints = len(jointIndices)  #self._pb.getNumJoints(bodyUniqueId)
    curPos, curOrn = self._pb.getBasePositionAndOrientation(bodyUniqueId)
    q1 = [curPos[0], curPos[1], curPos[2], curOrn[0], curOrn[1], curOrn[2], curOrn[3]]
    #print("q1=",q1)

    #qdot1 = [0,0,0, 0,0,0,0]
    baseLinVel, baseAngVel = self._pb.getBaseVelocity(bodyUniqueId)
    #print("baseLinVel=",baseLinVel)
    qdot1 = [
        baseLinVel[0], baseLinVel[1], baseLinVel[2], baseAngVel[0], baseAngVel[1], baseAngVel[2], 0
    ]
    #qError = [0,0,0, 0,0,0,0]
    desiredOrn = [
        desiredPositions[3], desiredPositions[4], desiredPositions[5], desiredPositions[6]
    ]
    axis1 = self._pb.getAxisDifferenceQuaternion(desiredOrn, curOrn)
    angDiff = [0, 0, 0]  #self.computeAngVel(curOrn, desiredOrn, 1, self._pb)
    qError = [
        desiredPositions[0] - curPos[0], desiredPositions[1] - curPos[1],
        desiredPositions[2] - curPos[2], angDiff[0], angDiff[1], angDiff[2], 0
    ]
    target_pos = np.array(desiredPositions)
    #np.savetxt("pb_target_pos.csv", target_pos, delimiter=",")

    qIndex = 7
    qdotIndex = 7
    zeroAccelerations = [0, 0, 0, 0, 0, 0, 0]
    useArray = True
    if useArray:
      jointStates = self._pb.getJointStatesMultiDof(bodyUniqueId,jointIndices)
    
    for i in range(numJoints):
      if useArray:
        js = jointStates[i]
      else:
        js = self._pb.getJointStateMultiDof(bodyUniqueId, jointIndices[i])

      jointPos = js[0]
      jointVel = js[1]
      q1 += jointPos

      if len(js[0]) == 1:
        desiredPos = desiredPositions[qIndex]

        qdiff = desiredPos - jointPos[0]
        qError.append(qdiff)
        zeroAccelerations.append(0.)
        qdot1 += jointVel
        qIndex += 1
        qdotIndex += 1
      if len(js[0]) == 4:
        desiredPos = [
            desiredPositions[qIndex], desiredPositions[qIndex + 1], desiredPositions[qIndex + 2],
            desiredPositions[qIndex + 3]
        ]
        #axis = self._pb.getAxisDifferenceQuaternion(desiredPos,jointPos)
        angDiff = self.computeAngVelRel(jointPos, desiredPos, 1, self._pb)
        #angDiff = self._pb.computeAngVelRel(jointPos, desiredPos, 1)

        jointVelNew = [jointVel[0], jointVel[1], jointVel[2], 0]
        qdot1 += jointVelNew
        qError.append(angDiff[0])
        qError.append(angDiff[1])
        qError.append(angDiff[2])
        qError.append(0)
        desiredVel = [
            desiredVelocities[qdotIndex], desiredVelocities[qdotIndex + 1],
            desiredVelocities[qdotIndex + 2]
        ]
        zeroAccelerations += [0., 0., 0., 0.]
        qIndex += 4
        qdotIndex += 4

    q = np.array(q1)

    qerr = np.array(qError)

    #np.savetxt("pb_qerro.csv",qerr,delimiter=",")

    #np.savetxt("pb_q.csv", q, delimiter=",")

    qdot = np.array(qdot1)
    #np.savetxt("qdot.csv", qdot, delimiter=",")

    qdotdesired = np.array(desiredVelocities)
    qdoterr = qdotdesired - qdot

    Kp = np.diagflat(kps)
    Kd = np.diagflat(kds)

    p = Kp.dot(qError)

    #np.savetxt("pb_qError.csv", qError, delimiter=",")
    #np.savetxt("pb_p.csv", p, delimiter=",")

    d = Kd.dot(qdoterr)

    #np.savetxt("pb_d.csv", d, delimiter=",")
    #np.savetxt("pbqdoterr.csv", qdoterr, delimiter=",")

    M1 = self._pb.calculateMassMatrix(bodyUniqueId, q1, flags=1)

    M2 = np.array(M1)
    #np.savetxt("M2.csv", M2, delimiter=",")

    M = (M2 + Kd * timeStep)

    #np.savetxt("pbM_tKd.csv",M, delimiter=",")

    c1 = self._pb.calculateInverseDynamics(bodyUniqueId, q1, qdot1, zeroAccelerations, flags=1)

    c = np.array(c1)
    #np.savetxt("pb_C.csv",c, delimiter=",")
    A = M
    #p = [0]*43
    #np.savetxt("pb_kp_dot_qError.csv", p)
    #np.savetxt("pb_kd_dot_vError.csv", d)

    b = p + d - c
    #np.savetxt("pb_b_acc.csv",b, delimiter=",")

    useNumpySolver = True
    if useNumpySolver:
      qddot = np.linalg.solve(A, b)
    else:
      qddot = self._pb.ldltSolve(bodyUniqueId, jointPositions=q1, b=b.tolist(), kd=kds, t=timeStep)

    tau = p + d - Kd.dot(qddot) * timeStep
    #print("len(tau)=",len(tau))
    #np.savetxt("pb_tau_not_clamped.csv", tau, delimiter=",")

    maxF = np.array(maxForces)
    #print("maxF",maxF)
    forces = np.clip(tau, -maxF, maxF)

    #np.savetxt("pb_tau_clamped.csv", tau, delimiter=",")
    return forces


class PDControllerStable(object):

  def __init__(self, pb):
    self._pb = pb

  def computePD(self, bodyUniqueId, jointIndices, desiredPositions, desiredVelocities, kps, kds,
                maxForces, timeStep):
    numBaseDofs = 0
    numPosBaseDofs = 0
    baseMass = self._pb.getDynamicsInfo(bodyUniqueId, -1)[0]
    curPos, curOrn = self._pb.getBasePositionAndOrientation(bodyUniqueId)
    q1 = []
    qdot1 = []
    zeroAccelerations = []
    qError = []
    if (baseMass > 0):
      numBaseDofs = 6
      numPosBaseDofs = 7
      q1 = [curPos[0], curPos[1], curPos[2], curOrn[0], curOrn[1], curOrn[2], curOrn[3]]
      qdot1 = [0] * numBaseDofs
      zeroAccelerations = [0] * numBaseDofs
      angDiff = [0, 0, 0]
      qError = [
          desiredPositions[0] - curPos[0], desiredPositions[1] - curPos[1],
          desiredPositions[2] - curPos[2], angDiff[0], angDiff[1], angDiff[2]
      ]
    numJoints = self._pb.getNumJoints(bodyUniqueId)
    jointStates = self._pb.getJointStates(bodyUniqueId, jointIndices)

    for i in range(numJoints):
      q1.append(jointStates[i][0])
      qdot1.append(jointStates[i][1])
      zeroAccelerations.append(0)
    q = np.array(q1)
    qdot = np.array(qdot1)
    qdes = np.array(desiredPositions)
    qdotdes = np.array(desiredVelocities)
    #qError = qdes - q
    for j in range(numJoints):
      qError.append(desiredPositions[j + numPosBaseDofs] - q1[j + numPosBaseDofs])
    #print("qError=",qError)
    qdotError = qdotdes - qdot
    Kp = np.diagflat(kps)
    Kd = np.diagflat(kds)
    p = Kp.dot(qError)
    d = Kd.dot(qdotError)
    forces = p + d

    M1 = self._pb.calculateMassMatrix(bodyUniqueId, q1)
    M2 = np.array(M1)
    M = (M2 + Kd * timeStep)
    c1 = self._pb.calculateInverseDynamics(bodyUniqueId, q1, qdot1, zeroAccelerations)
    c = np.array(c1)
    A = M
    b = -c + p + d
    qddot = np.linalg.solve(A, b)
    tau = p + d - Kd.dot(qddot) * timeStep
    maxF = np.array(maxForces)
    tau = np.clip(tau, -maxF, maxF)
    #print("c=",c)
    return tau



#add parent dir to find package. Only needed for source code build, pip install doesn't need it.
import os, inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(os.path.dirname(currentdir))
os.sys.path.insert(0, parentdir)

import gym
from pybullet_envs.bullet.racecarGymEnv import RacecarGymEnv

from baselines import deepq


def main():

  env = RacecarGymEnv(renders=True, isDiscrete=True)
  act = deepq.load("racecar_model.pkl")
  print(act)
  while True:
    obs, done = env.reset(), False
    print("===================================")
    print("obs")
    print(obs)
    episode_rew = 0
    while not done:
      env.render()
      obs, rew, done, _ = env.step(act(obs[None])[0])
      episode_rew += rew
    print("Episode reward", episode_rew)


if __name__ == '__main__':
  main()
